#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#

from abc import ABCMeta, abstractmethod


class SystemBase(metaclass=ABCMeta):
    """Base class for System component plugins"""

    def __init__(self, config):
        self.config = config

    @abstractmethod
    def configure(self):
        """Abstract class method to configure the System component"""

    @abstractmethod
    def build(self):
        """Abstract class method to build the System component"""



#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#

from roast.utils import get_combinations
import re


class TestCollect:
    def __init__(self):
        self.tests = {}

    def add_tests(self, **kwargs):
        combinations = self._get_test_combinations(**kwargs)
        self.tests.update(combinations)

    def remove_tests(self, **kwargs):
        combinations = self._get_test_combinations(**kwargs)
        for item in combinations.keys():
            self.tests.pop(item)

    def test_map(self, tests_list, parameters, scenario=[]):
        for test in tests_list:
            test = [test]
            test.extend(scenario)
            self.tests[tuple(test)] = parameters

    def _get_test_combinations(self, **kwargs):
        keys = kwargs.keys()
        values = list(kwargs.values())
        combinations = {}
        # FIXME: Empty list is not handled
        for combination in get_combinations(values):
            key = "-".join([myval for myval in combination if myval])
            combinations[key] = dict(zip(keys, combination))
        return combinations

    def get_tests(self, silent_discard=False, sort=False):
        ret = list(self.tests.keys())
        if not ret and not silent_discard:
            raise Exception("ERROR: Empty tests list")
        if sort:
            return sorted(ret)
        else:
            return ret

    def get_test_params(self, test_name):
        return self.tests[test_name]  # returns dictinary with key value pairs



#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#

from abc import ABCMeta, abstractmethod


class TestSuiteBase(metaclass=ABCMeta):
    """Base class for Test Suite component plugin"""

    def __init__(self, config):
        self.config = config

    @abstractmethod
    def configure(self):
        """Abstract class method to configure the TestSuite component"""

    @abstractmethod
    def build(self):
        """Abstract class method to build the TestSuite component"""

    @abstractmethod
    def deploy(self):
        """Abstract class method to deploy the TestSuite component"""

    @abstractmethod
    def run(self):
        """Abstract class method to run the TestSuite component"""



#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#

import logging
import atexit
from roast.xexpect import Xexpect
from abc import ABCMeta, abstractmethod
from stevedore import driver

log = logging.getLogger(__name__)


class Serial:
    def __init__(self, serial_type: str, config, **kwargs) -> None:
        self._serial_mgr = driver.DriverManager(
            namespace="roast.serial",
            name=serial_type,
            invoke_on_load=True,
            invoke_args=(config,),
            invoke_kwds=kwargs,
        )

    @property
    def driver(self):
        return self._serial_mgr.driver

    def exit(self) -> None:
        self.driver.exit()


class SerialBase(metaclass=ABCMeta):
    def __init__(self, config) -> None:
        self.config = config
        self.hostname = ""
        self.is_live = False
        self._configure()
        self.console = Xexpect(log, hostname=self.hostname, non_interactive=False)
        self.expect = self.console.expect
        self.sendline = self.console.sendline
        self.runcmd = self.console.runcmd
        self.runcmd_list = self.console.runcmd_list
        self.sendcontrol = self.console.sendcontrol
        self.send = self.console.send
        self.output = self.console.output
        self._setup_init = self.console._setup_init
        self.search = self.console.search
        self.sync = self.console.sync
        atexit.register(self.exit)
        self._connect()
        self.is_live = True

    def _configure(self) -> None:
        """Set hostname."""

    @abstractmethod
    def _connect(self) -> None:
        """Connect to host."""

    @abstractmethod
    def exit(self) -> None:
        """Release connection."""

    @property
    def prompt(self):
        return self.console.prompt

    @prompt.setter
    def prompt(self, value):
        self.console.prompt = value



#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#

import inspect
import logging
import time
import numpy as np
from enum import Enum
from collections import Counter, defaultdict
from typing import Any, List, Type, Sequence
from mimesis import BaseDataProvider, BaseProvider, Choice, Development
from roast.providers.choices import Choices
from roast.providers.shuffle import Shuffle
from roast.utils import read_json, write_json, frange, mkfile, to_nested
from box import Box, BoxKeyError
from plusminus import ArithmeticParser
import pyparsing as pp
from roast.exceptions import RandomizerError

try:
    # For python 3.8 and later
    from mimesis.types import Seed
except ImportError:
    # For everyone else
    from mimesis.typing import Seed

__all__ = ["Randomizer"]

log = logging.getLogger(__name__)


class WeightPreset(Enum):
    LOW_HEAVY = 0
    HIGH_HEAVY = 1
    NORMAL = 2
    INVERSE_NORMAL = 3
    EXTREME_LIMITS = 4


class Randomizer(BaseDataProvider):
    def __init__(self, seed: Seed = None, randomize: bool = True) -> None:
        """Class that contains core Xilinx randomization data providers - `choice`,
        `choices`, `shuffle`, and `boolean`.

        The methods included provide complex randomization functionality based on core providers.

        A `data` attribute holds values randomly generated.

        Args:
            seed (Seed, optional): Seed for random. When set to `None` the current system time is
                used. Defaults to None.
            randomize (bool, optional): Retrieve randomized or default value. Defaults to True.
        """

        if seed is None:
            seed = int(time.time() * 1000)  # system time
        super().__init__(seed=seed)
        self.choice = Choice(seed=self.seed)
        self.choices = Choices(seed=self.seed)
        self.shuffle = Shuffle(seed=self.seed)
        self.boolean = Development(seed=self.seed).boolean
        self.randomize = randomize
        self.parser = ArithmeticParser()
        self.evaluate = self.parser.evaluate
        log.debug(f"seed={self.seed}")
        self.data = Box(
            conversion_box=False, default_box=True, default_box_attr=[], box_dots=True
        )
        self.replace = True
        self._excludes_file = ""
        self.rng = np.random.default_rng(seed)
        self.weights = defaultdict(list)

    @property
    def datafile(self):
        return self._datafile

    @datafile.setter
    def datafile(self, filename):
        self._datafile = filename
        _parameters = read_json(self._datafile, decode_numbers=True)
        self.parameters = Box(_parameters, box_dots=True)

    @property
    def excludes_file(self):
        return self._excludes_file

    @excludes_file.setter
    def excludes_file(self, filename):
        self._excludes_file = filename
        try:
            _data = read_json(filename, decode_numbers=True)
        except:
            mkfile(filename)
            _data = {}
        self.data = Box(
            _data,
            conversion_box=False,
            default_box=True,
            default_box_attr=[],
            box_dots=True,
        )
        self.replace = False

    def reset_excluded(self, attribute: str = "") -> None:
        if not attribute:
            self.data = {}
        elif not self.excludes_file:
            self.data[attribute] = []
        else:
            try:
                _json = read_json(self.excludes_file, decode_numbers=True)
            except ValueError:
                _json = {}
            _data = Box(_json, box_dots=True)
            if _data:
                _data[attribute] = []
            self.data[attribute] = []
        self.to_json()

    def to_json(self, filename="") -> None:
        """Dumps `data` attribute to JSON file. If filename is not specified, the filename set the
        through the excludes_file method will be used.

        Args:
            filename (str, optional): Filename where JSON data will be written. Defaults to "".
        """
        if filename:
            write_json(filename, to_nested(self.data))
        elif self.excludes_file:
            write_json(self.excludes_file, to_nested(self.data))

    def generate_weights(
        self, length, a: int = 1, b: int = 1, preset: WeightPreset = None
    ) -> Sequence[float]:
        """Generates a histogram over length-1 bins defined by the Beta probability distribution
        function. If defaults values are used, this is equivalent to a uniform distribution.

        Args:
            length (int): Number of weights
            a (int, optional): Alpha shape parameter. Defaults to 1.
            b (int, optional): Beta shape parameter. Defaults to 1.
            preset (WeightPreset, optional): Predefined Alpha and Beta values. Defaults to None.

        Returns:
            Sequence[float]: Relative weights.
        """
        if preset == WeightPreset.LOW_HEAVY:
            a = 1
            b = 10
        elif preset == WeightPreset.HIGH_HEAVY:
            a = 10
            b = 1
        elif preset == WeightPreset.NORMAL:
            a = 5
            b = 5
        elif preset == WeightPreset.INVERSE_NORMAL:
            a = 0.5
            b = 0.5
        elif preset == WeightPreset.EXTREME_LIMITS:
            a = 0.1
            b = 0.1
        try:
            p = self.rng.beta(a, b, length * 100)  # generate elements over distribution
            s = self.rng.binomial(length - 1, p)  # divide into bins
            c = Counter(s)  # count number of occurrences for each bin
            keys, weights = zip(
                *sorted(c.items())
            )  # sort ascending and extract bin number and weight
            if len(c) != length:  # interpolate if any bins are missing
                weights = np.interp(range(length), keys, weights)
        except MemoryError as e:
            log.error(e)
            raise
        return weights

    def check_attribute(self, attribute: str) -> None:
        """Checks if attribute exists in parameters.

        Args:
            attribute (str): Attribute name.

        Raises:
            KeyError: Exception raised if attribute not found.
        """
        try:
            self.parameters[attribute]
        except BoxKeyError:
            err_msg = f"Attribute {attribute} not defined."
            log.error(err_msg)
            raise KeyError(err_msg)

    def generate_attribute_weights(
        self, attribute: str, a: float = 1, b: float = 1, preset: WeightPreset = None
    ) -> None:
        """Generates weights for attribute and cached in memory to avoid regeneration.

        Args:
            attribute (str): Attribute name.
            a (int, optional): Alpha shape parameter. Defaults to 1.
            b (int, optional): Beta shape parameter. Defaults to 1.
            preset (WeightPreset, optional): Predefined Alpha and Beta values. Defaults to None.
        """
        excluded = self.get_excluded(attribute)
        elements = self.get_elements(attribute)
        weights = list(self.generate_weights(len(elements), a, b, preset))
        indices = sorted(
            [elements.index(exclude_element) for exclude_element in excluded]
        )
        for index in indices:
            del weights[index]
        self.weights[attribute] = weights

    def get_excluded(self, attribute: str) -> Sequence[Any]:
        """If global replacement or parameter replamcent is disabled, construct
        the sequence of elements to be excluded from selection. The parameter setting
        always overrides the global setting.

        Args:
            attribute (str): Attribute name.

        Returns:
            Sequence[Any]: Elements to be excluded from selection.
        """
        self.check_attribute(attribute)
        param_replace = self.parameters[attribute].get("replace")
        try:
            if (self.replace and param_replace is None) or param_replace:
                excluded = self.parameters[f"{attribute}.excluded"]
            else:
                excluded = list(self.parameters[f"{attribute}.excluded"]) + list(
                    self.data[attribute]
                )
        except KeyError:
            if (self.replace and param_replace is None) or param_replace:
                excluded = []
            else:
                excluded = self.data[attribute]
        return excluded

    def get_elements(self, attribute: str) -> List[Any]:
        """Constructs list of elements for possible selection. Either elements or range must
        exist or an exception will be thrown. If both exist, elements has precedence.

        Args:
            attribute (str): Attribute name.

        Raises:
            KeyError: Raised when neither elements nor range is defined.

        Returns:
            Sequence[Any]: List of elements for possible selection.
        """
        self.check_attribute(attribute)
        if "elements" in self.parameters[attribute]:
            elements = self.parameters[f"{attribute}.elements"]
        elif "range" in self.parameters[attribute]:
            elements = list(frange(*self.parameters[f"{attribute}.range"]))
        else:
            err_msg = f"For attribute {attribute}, elements or range not defined."
            log.error(err_msg)
            raise KeyError(err_msg)
        return elements

    def get_value(
        self, attribute: str, store_data: bool = True, preset=None, **kwargs
    ) -> Any:
        """Gets a randomized value defined by attribute name. Possible values can be discrete
        values or range with excluded values removed.

        Args:
            attribute (str): Dot-based key name to retrieve value.
            store_data (bool, optional): Store randomized value to `data` attribute. Defaults to
            True.
            preset (WeightPreset, optional): Predefined Alpha and Beta values. Defaults to None.
            kwargs: Keyword arguments for shape parameters.

        Returns:
            Random choice of possible elements if randomize is True.
            Default value if randomize is False.

        Raises:
            KeyError: If `attribute` is not found in parameters.

        """
        self.check_attribute(attribute)
        param_replace = self.parameters[attribute].get("replace")
        if self.randomize and self.parameters[attribute].get("randomize", True):
            elements = self.get_all_values(attribute)
            # if preset or kwargs, generate weights for attribute
            if preset or "preset" in self.parameters[attribute]:
                if "preset" in self.parameters[attribute]:
                    param_preset = self.parameters[f"{attribute}.preset"]
                    preset = WeightPreset[param_preset]
                if not self.weights[attribute]:
                    self.generate_attribute_weights(attribute, preset=preset)
                element = self.choices(elements, self.weights[attribute])[0]
            elif kwargs or "shape" in self.parameters[attribute]:
                if "shape" in self.parameters[attribute]:
                    shape = self.parameters[f"{attribute}.shape"]
                    kwargs["a"] = shape[0]
                    kwargs["b"] = shape[1]
                if not self.weights[attribute]:
                    self.generate_attribute_weights(attribute, **kwargs)
                element = self.choices(elements, self.weights[attribute])[0]
            else:
                element = self.choice(elements)
            if store_data and element not in self.data[attribute]:
                self.data[attribute].append(element)
            # if excluding previous and weights have been generated, delete weight for selected element
            if (self.replace and param_replace is None) or param_replace:
                return element
            else:
                if preset or kwargs:
                    index = elements.index(element)
                    del self.weights[attribute][index]
            return element
        else:
            try:
                format = self.parameters[attribute].get("format")
                element = self.parameters[f"{attribute}.default"]
                if format:
                    return f"{element:{format}}"
                else:
                    return element
            except BoxKeyError:
                err_msg = f"Default value for {attribute} not defined."
                log.error(err_msg)
                raise KeyError(err_msg)

    def get_all_values(self, attribute: str) -> Sequence[Any]:
        """Gets all values defined by attribute name. Possible values can be discrete
        values or range with excluded values removed.

        Args:
            attribute (str): Dot-based key name to retrieve all values.

        Returns:
            List of all possible values with excluded values removed.

        Raises:
            KeyError: If `attribute` is not found in parameters.

        """
        excluded = self.get_excluded(attribute)
        elements = self.get_elements(attribute)
        elements = [element for element in elements if element not in excluded]
        format = self.parameters[attribute].get("format")
        if format:
            elements = [f"{element:{format}}" for element in elements]
        return elements

    def generate_sequence(self, expr, attrs, max_tries=200, **kwargs) -> dict:
        """Generate a sequence of values defined by expression in attribute order. Two variables,
        'prev' and 'current', are included to simplify the comparison. In order for the "current"
        value to be added, it must satisfy the expression condition compared to the "previous"
        value. However, any expression can be used to define the condition.

        Args:
            expr (str): Expression to determine whether value is added to output sequence.
            attrs (:obj:`list` of :obj:`str`): Attributes to retrieve randomized values.
            max_tries (int, optional): Number of tries to generate random values. Defaults to 200.
            **kwargs: Keyword arguments to be declared into expression evaluation.

        Returns:
            dict: Randomized values defined by expression.

        Raises:
            Exception: If max_tries is reached without generating valid sequence.

        """
        expr_underscore = expr.replace(
            ".", "__"
        )  # Convert dot-based keys to double underscore
        _data = {}

        # Push kwargs to evaluate engine
        for key, value in kwargs.items():
            self.evaluate(f"{key} = {value}")

        for i in range(max_tries):
            try:
                for j, attr in enumerate(attrs):
                    if j == 0:
                        _data[attr] = self.get_value(attr, store_data=False)
                        attr_underscore = attr.replace(".", "__")
                        self.evaluate(f"{attr_underscore} = {_data[attr]}")
                        log.debug(f"i={i}, j={j}, data={_data}")
                    else:
                        for k in range(len(self.get_all_values(attr))):
                            _data[attr] = self.get_value(attr, store_data=False)
                            attr_underscore = attr.replace(".", "__")
                            self.evaluate(f"{attr_underscore} = {_data[attr]}")
                            log.debug(f"i={i}, j={j}, k={k}, data={_data}")
                            prev = _data[attrs[j - 1]]
                            self.evaluate(f"prev = {prev}")
                            current = _data[attr]
                            self.evaluate(f"current = {current}")
                            if self.evaluate(expr_underscore):
                                break
                        else:
                            raise Exception(f"Data generation failed for: {attr}")
                log.info("Data generation successful.")
                break
            except Exception as e:
                log.warning(e)
                continue
        else:
            raise RandomizerError(f"Data generation failed.")

        for key, value in _data.items():
            if value not in self.data[key]:
                self.data[key].append(value)
        return _data

    def generate_conditional(self, expr, max_tries=200, **kwargs) -> dict:
        """Generate randomized values defined by expression.

        Args:
            expr (str): Expression to determine whether value is added to output sequence.
            max_tries (int, optional): Number of tries to generate random values. Defaults to 200.
            **kwargs: Keyword arguments to be declared into expression evaluation.

        Returns:
            dict: Randomized values defined by expression.

        Raises:
            Exception: If max_tries is reached without generating valid sequence.

        """
        expr_underscore = expr.replace(
            ".", "__"
        )  # Convert dot-based keys to double underscore
        _data = {}

        # Push kwargs to evaluate engine
        for key, value in kwargs.items():
            self.evaluate(f"{key} = {value}")

        for i in range(max_tries):
            try:
                if self.evaluate(expr_underscore):
                    log.debug(f"expr: {expr}")
                    log.info("Data generation successful.")
                    self._clear_variables(_data)
                    break
                else:
                    raise Exception(f"Generated data failed condition: {expr}")
            except NameError as e:  # Missing variable is randomized attribute
                qs = pp.QuotedString("'")
                attr_underscore = qs.searchString(e)[0][0]
                attr = attr_underscore.replace("__", ".")
                _data[attr] = self.get_value(attr, store_data=False)
                self.evaluate(f"{attr_underscore} = {_data[attr]}")
                log.debug(f"i={i}, data={_data}")
            except Exception as e:
                log.warning(e)
                self._clear_variables(_data)
                continue
        else:
            raise RandomizerError(f"Data generation failed.")

        for key, value in _data.items():
            if value not in self.data[key]:
                self.data[key].append(value)
        return _data

    def _clear_variables(self, _data):
        for key in _data.keys():
            key_underscore = key.replace(".", "__")
            self.evaluate(f"{key_underscore} =")

    class Meta:
        """Class for metadata."""

        name = "randomizer"

    def __dir__(self) -> List[str]:
        """Available data providers.

        The list of result will be used in AbstractField to
        determine method's class.

        :return: List of attributes.
        """
        attributes = []
        exclude = BaseDataProvider().__dict__.keys()

        for a in self.__dict__:
            if a not in exclude:
                if a.startswith("_"):
                    attribute = a.replace("_", "", 1)
                    attributes.append(attribute)
                else:
                    attributes.append(a)
        return attributes

    def add_provider(self, cls: Type[BaseProvider]) -> None:
        """Add a custom provider to Generic() object.

        :param cls: Custom provider.
        :return: None
        :raises TypeError: if cls is not class or is not a subclass
            of BaseProvider.
        """
        if inspect.isclass(cls):
            if not issubclass(cls, BaseProvider):
                raise TypeError("The provider must be a " "subclass of BaseProvider")
            try:
                meta = getattr(cls, "Meta")
                name = getattr(meta, "name")
            except AttributeError:
                name = cls.__name__.lower()
            setattr(self, name, cls(seed=self.seed))
        else:
            raise TypeError("The provider must be a class")

    def add_providers(self, *providers: Type[BaseProvider]) -> None:
        """Add a lot of custom providers to Generic() object.

        :param providers: Custom providers.
        :return: None
        """
        for provider in providers:
            self.add_provider(provider)



#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#

import os
import sys
import git
import glob
import logging
import fileinput
import shutil
import inspect
import argparse
import re
import random
import string
from filelock import FileLock
import itertools, collections.abc
import yaml
import subprocess
import atexit
from pathlib import Path
from distutils.dir_util import copy_tree
from git.exc import GitCommandError, InvalidGitRepositoryError
from roast.exceptions import DirectoryNotFoundError, GitError

from typing import Any, List, Optional, Dict
from warnings import warn

log = logging.getLogger(__name__)


class Git:
    def __init__(self, git_params, repo_path, clone_once):
        self.repo = None
        self.repo_url = git_params.url
        self.branch = git_params.branch
        self.sparse_path = git_params.get("sparse_path", [])
        self.patches = git_params.get("patches", [])
        self.patchesDir = git_params.get("patchesDir")
        self.rev = git_params.get("rev")
        self.repo_path = repo_path
        self.clone_once = clone_once

    def clone(self, **kwargs):
        self.repo_name = get_base_name(self.repo_path)
        self.base_path = get_dir_name(self.repo_path)
        self.lock_path = os.path.join(self.base_path, f".{self.repo_name}.lock")
        self.git_path = os.path.join(self.repo_path, ".git")
        self.repo_cloned = False
        self.lock = FileLock(self.lock_path)
        with self.lock:
            if (self.clone_once and not is_dir(self.git_path)) or not self.clone_once:
                self._clone(**kwargs)
                self.repo_cloned = True

    def _clone(self, **kwargs):

        log.info(f"Cloning {self.repo_url}")
        log.info("******************************************************")
        log.info(f"repo_url : {self.repo_url}")
        log.info(f"branch : {self.branch}")
        log.info(f"sparse_path : {self.sparse_path}")
        log.info(f"patches : {self.patches}")
        log.info(f"patchesDir : {self.patchesDir}")
        log.info(f"rev : {self.rev}")
        log.info(f"repo_path : {self.repo_path}")
        log.info("******************************************************")
        no_checkout = True if len(self.sparse_path) else False
        git.Repo.clone_from(
            self.repo_url,
            self.repo_path,
            branch=self.branch,
            no_checkout=no_checkout,
            **kwargs,
        )

        self.repo = git.Repo(self.repo_path)
        if len(self.sparse_path):
            with self.repo.config_writer() as cw:
                cw.set_value("core", "sparseCheckout", "true")
            checkout_path = os.path.join(self.repo_path, ".git/info/sparse-checkout")
            with open(checkout_path, "w") as fd:
                fd.writelines(self.sparse_path)
            self.repo.git.checkout(self.branch)

        if kwargs.get("recurse_submodules") == "latest":
            for submodule in self.repo.submodules:
                submodule.update(init=True)
                sub_repo = submodule.module()
                sub_repo.git.pull()

    def checkout(self):
        if self.rev:
            self.repo.git.checkout(self.rev)

    def apply_patch(self, workDir):

        self.repo = git.Repo(self.repo_path)
        patches = []
        if self.patches and is_dir(f"{workDir}/patches"):
            patches = [
                os.path.join(f"{workDir}/patches", patch) for patch in self.patches
            ]

        if is_dir(self.patchesDir):
            external_patches = get_files(
                self.patchesDir, extension="patch", abs_path=True
            )
            patches.extend(external_patches)

        for patch in patches:
            patch_file_name = get_base_name(patch)

            # Exclude coverletter patches if present.
            if is_file(patch) and not patch_file_name.startswith("0000-"):
                self.repo.git.am(["-3", patch])

    def clean(self):
        self.repo.git.clean("-xdf")

    def run_gitcmd(self, cmd, ret_out=False):
        """Run the git command specified by the cmd input, in the specified repo
        path.

        Args:
            cmd (str): the git command to be executed
            ret_out (bool, optional): decides if the output of git commands need
                     to be returned for further processing. Defaults to False.

        Returns:
            boolean/list: pass fail status of git command/ [status,
                            output,error] from the command
        """
        g = git.cmd.Git(os.path.realpath(self.repo_path))
        cmd_list = cmd.split()
        stat, out, err = g.execute(cmd_list, with_extended_output=True)
        if ret_out:
            return stat, out, err
        return stat

    def git_update_repo(self, branch=None, commit=None, tag=None, ret_out=False):
        """Updates the user provided repo to user specified tag or branch
            or (branch and optional commit ID)
            Note: if no tag, branch, commit is provided it updates the repo to
            its latest master branch

        Args:
            branch (str, optional): branch name to check out. Defaults to None.
            commit (str, optional): commit sha to check out, to be used with
                    valid branch. Defaults to None.
            tag (str, optional): tag that needs to be checked out. Defaults to None.
            ret_out (bool, optional): boolean that decides if deatiled output of
                    updating need to be returned. Defaults to False.

        Returns:
            boolean/list: pass fail status of repo update/ [status, output]
                            list from the command
        """
        status = False
        out = ""
        if not self.repo_path:
            out += "No repo path provided to update"
            return status, out
        if not is_git_repo(self.repo_path):
            out += f"repo path: {self.repo_path} is not a git repository"
            return status, out

        out += f"Updating repository @ {self.repo_path}\n"
        # Get current git repo state
        gp = self.repo_path
        repo = git.Repo(gp)
        git_commit = repo.git.rev_parse(repo.head.commit.hexsha, short=7)
        try:
            git_branch = repo.active_branch.name
        except TypeError:
            git_branch = None
        try:
            _, git_tag, _ = self.run_gitcmd(
                f"git describe --exact-match {git_commit}", True
            )
        except GitCommandError:
            git_tag = None
        out += f"Repo at branch: {git_branch} commit: {git_commit} tag: {git_tag}\n"
        out += f"Requested branch: {branch} commit: {commit} tag: {tag}\n"
        # Check if the repo is already in requested state
        if tag:
            if git_tag == tag:
                out += f"Not updating, repo already at required tag : {tag}"
                log.debug(out)
                if ret_out:
                    return True, out
                else:
                    return
        if branch and commit:
            short_sha = repo.git.rev_parse(commit, short=7)
            if git_commit == short_sha:
                out += f"Not updating, repo already at commit: {git_commit}"
                log.debug(out)
                if ret_out:
                    return True, out
                else:
                    return
        # Update the repo to user specified state
        status = status or self.run_gitcmd("git remote update -p")
        status = status or self.run_gitcmd("git reset --hard")
        if tag:
            status = status or self.run_gitcmd(f"git checkout {tag}")
            out += f"checked out tag {tag}"
        elif branch:
            status = status or self.run_gitcmd(f"git checkout {branch}")
            status = status or self.run_gitcmd(f"git pull origin {branch}")
            out += f"updated branch: {repo.active_branch.name}"
            if commit:
                status = status or self.run_gitcmd(f"git checkout {commit}")
                out += f" checked out commit {commit}"
        else:  # No tag/branch provided, update to latest master branch
            status = status or self.run_gitcmd("git checkout master")
            status = status or self.run_gitcmd("git pull origin master")
            commit = repo.git.rev_parse(repo.head.commit.hexsha, short=7)
            out += f"updated to master branch, commit {commit}"
        log.debug(out)
        if ret_out:
            return not status, out
        return not status

    def log(self):
        # print log for a valid git repo
        if is_git_repo(self.repo_path):
            self.repo = git.Repo(self.repo_path)
            # repo_url = self.repo.git.remote("get-url", "origin")
            # FIXME due to git version not supporting the command (Required git version 2.23.0)
            remote = self.repo.git.remote()
            repo_url = self.repo.git.config("--get", f"remote.{remote}.url")
            repo_current_branch = self.repo.git.rev_parse(["--abbrev-ref", "HEAD"])
            self.local_commits = ""
            self.origin_commit = ""
            if repo_current_branch != "HEAD":
                repo_branch_str = self.repo.git.log(
                    "--pretty=format:%d",
                )
                # Get all the remote branches
                remote_branches = [
                    branch
                    for branch in list(filter(None, repo_branch_str.split("\n")))
                    if remote in branch
                ]
                # Get the latest remote branch
                remote_branches = remote_branches[0].split(",")
                for branch in remote_branches:
                    branch = re.sub("[() ]", "", branch)
                    if remote in branch and "HEAD" not in branch:
                        remote_branch = branch
                self.local_commits = self.repo.git.log(
                    f"{remote_branch}...{repo_current_branch}",
                    '--pretty=format:"%H - %an, %ar : %s"',
                )
                self.origin_commit = self.repo.git.log(
                    f"{remote_branch}",
                    "-n 1",
                    '--pretty=format:"%H - %an, %ar : %s"',
                )
            else:
                repo_current_branch = self.repo.git.branch()
                match_obj = re.search(
                    "HEAD detached (at|from) ([a-z0-9]+)", repo_current_branch
                )
                if match_obj:
                    repo_branch_commit = match_obj.group(2)
                    repo_head_commit = self.repo.git.rev_parse(["--verify", "HEAD"])
                    self.local_commits = self.repo.git.log(
                        f"{repo_branch_commit}...{repo_head_commit}",
                        '--pretty=format:"%H - %an, %ar : %s"',
                    )
                    self.origin_commit = self.repo.git.log(
                        "-n 1",
                        repo_branch_commit,
                        '--pretty=format:"%H - %an, %ar : %s"',
                    )
                if not self.origin_commit:
                    remote = "HEAD"
                    self.origin_commit = self.repo.git.log(
                        "-n 1", "--pretty=format:%H - %an, %ar : %s"
                    )

            log.info("******************************************************")
            log.info(f" REPO URL: {repo_url}")
            log.info(f" BRANCH: {repo_current_branch}")
            if self.local_commits:
                log.info(f" LOCAL_COMMITS: {self.local_commits}")
            log.info(f" {remote.upper()} COMMIT: {self.origin_commit}")
            log.info("******************************************************")


def is_git_repo(repo_path):
    """Checks if a given path is a git repo

    Args:
        repo_path (os.path): path to the dir that needs to be tested to see if the dir is a git repo

    Returns:
        bool: True if path is a git repo.
    """

    try:
        _ = git.Repo(repo_path).git_dir
        return True
    except InvalidGitRepositoryError:
        return False


def clone(git_params, repo_path, workDir=None, clone_once=False, **kwargs):

    git = Git(git_params, repo_path, clone_once)

    git.clone(**kwargs)

    if git.repo_cloned:

        git.checkout()

        git.apply_patch(workDir)

    git.log()


def surround_double_quotes(string: str) -> str:
    """Surrounds the input string with double quotes and returns the modified string.

    Args:
        string: String to be modified.

    Returns:
        str: String with double quotes.
    """

    return '"' + str(string) + '"'


def print_err_exit(err_msg: str) -> None:
    """Print the Error Message and Exits the Program.

    Args:
        err_msg: Error Message.
    """

    print("ERROR: " + err_msg)

    sys.exit(1)


def print_warn(warn_msg: str) -> None:
    """Print the Warning Message and continues the Program execution.

    Args:
        warn_msg: Warning Message.
    """
    print("WARNING: " + warn_msg)


def print_msg(msg: str) -> None:
    """Print * across the screen width and puts the msg in the center of the screen.

    Args:
        msg: Message to be print.
    """
    try:
        _, columns = os.popen("stty size", "r").read().split()
        print(str(msg).center(int(columns), "*"))
    except:
        print(str(msg))


def is_var_type(var: Any, expected_type: Any) -> bool:
    """Returns True if the expected variable type and provided variable type matches

    Args:
        var: Any valid python variable.
        expected_type: Any valid python variable type.

    Returns:
        bool: Returns True, else Exits the Program.
    """

    if not (type(var) == expected_type):
        msg = "Expected:" + str(expected_type) + " " + "Received:" + str(type(var))
        print_err_exit(msg)
    return True


def stripper(elements_list: List[str]) -> List[str]:
    """Strips unwanted spaces from start and end, in a list of strings.

    Args:
        elements_list: List of strings.

    Returns:
        :obj:`list` of :obj:`str`: Stripped strings list.
    """
    is_var_type(elements_list, list)
    return [element.strip() for element in elements_list]


def is_defined(args, mandatory_args):
    """Checks if the argparser.args contains all the mandatory args, Else report error and exit.

    Args:
        args (:obj:`ArgParser.args`): Argparser args.
        mandatory_args: Mandatory args.
    """

    # Test for the mandatory args
    for arg in mandatory_args:
        if arg not in args.keys() or args[arg] is None or args[arg] == "":
            msg = surround_double_quotes(arg) + " " + "is not defined"
            print_err_exit(msg)


def del_element(
    element: str, elements_list: List[str], silent_discard: bool = True
) -> None:
    """Removes an element from a list.

    Args:
        element: Element to be removed.
        elements_list: List of Strings.
        silent_discard: Discard Message. Defaults to True.
    """

    is_var_type(elements_list, list)

    try:
        elements_list.remove(element)
    except:
        if not silent_discard:
            msg = (
                "KeyError:" + " " + surround_double_quotes(element) + " " + "not found"
            )
            print_err_exit(msg)


def write_to_yaml(data: str, filepath: str, silent_discard: bool = True) -> None:
    """Write the data to YAML file, raises Error Message on demand.

    Args:
        data: Data to be written.
        filepath: Path of file.
        silent_discard: Discard Message. Defaults to True.
    """

    is_successful = False
    try:
        with open(filepath, "w") as outfile:
            yaml.dump(data, outfile, default_flow_style=False)
            is_successful = True
    except:
        pass
    if not silent_discard:
        if is_successful:
            print("Write Data to file: %s is successful" % filepath)
        else:
            msg = (
                "Write Data to file:"
                + " "
                + surround_double_quotes(filepath)
                + " "
                + "failed"
            )
            print_err_exit(msg)


def del_key(key: str, dictionary: dict, silent_discard: bool = True) -> None:
    """Deletes Key from a dictionary, and raises Not Found Error Message on demand.

    Args:
        key: Key to be deleted.
        dictionary: Dictionary.
        silent_discard: Discard Message. Defaults to True.
    """

    is_var_type(dictionary, dict)

    try:
        del dictionary[key]
    except:
        if not silent_discard:
            msg = "KeyError:" + " " + surround_double_quotes(key) + " " + "not found"
            print_err_exit(msg)


def get_env(env_var_list: List) -> dict:
    """Return dictionary of the env variables Else raises Error Message.

    Args:
        env_var_list: List of the variables expected to be present in environment.

    Returns:
        dict: Environment variables.
    """

    is_var_type(env_var_list, list)

    if len(env_var_list) <= 0:
        msg = "Empty env_var_list"
        print_err_exit(msg)

    vars_dict = {}
    for var in env_var_list:
        if os.environ.get(var) is None:
            msg = (
                "ENV_VAR:" + " " + surround_double_quotes(var) + " " + "is not defined"
            )
            print_err_exit(msg)
        else:
            vars_dict[var] = os.environ.get(var)
    return vars_dict


def is_file(filepath: str, silent_discard: bool = True) -> bool:
    """Return True if the file exists Else returns False and raises Not Found Error Message.

    Args:
        filepath: File Path.

    Raises:
        FileNotFoundError: Raises exception if file not found.

    Returns:
        bool: True, if file is found Or False, if file is not found.
    """

    try:
        if os.path.isfile(filepath):
            return True
        else:
            raise Exception
    except:
        if not silent_discard:
            err_msg = f"No such file exists: {filepath}"
            log.exception(err_msg)
            raise FileNotFoundError(err_msg) from None
        return False


def is_dir(dirpath: str, silent_discard: bool = True) -> bool:
    """Checks if directory exists.

    Args:
        dirpath: Directory Path.

    Raises:
        DirectoryNotFoundError (Exception): Raises exception if directory not found.

    Returns:
        bool: True, if directory is found Or False, if directory is not found.
    """

    try:
        if os.path.isdir(dirpath):
            return True
        else:
            raise Exception
    except:
        if not silent_discard:
            err_msg = f"No such directory exists: {dirpath}"
            raise DirectoryNotFoundError(err_msg, log_stack=True) from None
        return False


def remove(path: str, silent_discard: bool = True) -> None:
    """Removes any file or folder recursively, if it exists else reports error message based on user demand.

    Args:
        path: Directory or file path.
    """
    is_successful = False
    try:
        if is_dir(path):
            os.sync()
            shutil.rmtree(path)
            is_successful = True
        else:
            os.remove(path)
            is_successful = True
    except OSError as e:
        pass
    if not silent_discard:
        if is_successful:
            print("%s Removed successfully" % path)
        else:
            print("Error: %s" % path, e.strerror)  # FIXME: e doesn't exist here
            sys.exit(1)


def reset(path: str) -> None:
    remove(path)
    mkdir(path)


def load_yaml(filepath: str) -> Optional[dict]:
    """Read yaml file data and returns data in a dict format.

    Args:
        filepath: Path of the yaml file.

    Returns:
        dict: Return Python dict if the file reading is successful.
    """

    if is_file(filepath):
        try:
            with open(filepath) as f:
                data = yaml.safe_load(f)
            return data
        except:
            print("%s file reading failed" % filepath)
            return None
    else:
        return None


def mkdir(folderpath: str, silent_discard: bool = True) -> None:
    """Create the folder structure, raises Error Message on demand.

    Args:
        folderpath: Path of the folder structure.
    """
    is_successful = False
    try:
        os.makedirs(folderpath)
        is_successful = True
    except:
        pass

    if not silent_discard:
        if is_successful:
            print("%s Directory created " % folderpath)
        else:
            print("%s Unable to create directory " % folderpath)
            sys.exit(1)


def mkfile(fname: str, silent_discard: bool = True) -> None:
    """Open an empty file, raises Error Message on demand
       Equivalent to touch in shell

    Arguments:
        filepath {string} -- Path of file
    """
    is_successful = False
    try:
        Path(fname).touch()
        is_successful = True
    except:
        pass
    if not silent_discard:
        if is_successful:
            print("%s Created File" % fname)
        else:
            print("%s Unable to create file " % fname)
            sys.exit(1)


def is_error(stderr_msg: str) -> bool:
    """Return True if the error string not found else exits with error message.

    Args:
        stderr_msg: std error message.
    """
    err_list = ["fatal", "error", "Error:"]

    for err in err_list:
        if err in stderr_msg:
            print_err_exit(stderr_msg)
    return True


def detailed_error():
    exc_type, _, exc_tb = sys.exc_info()
    fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
    print(exc_type, fname, exc_tb.tb_lineno)


def copy_file(src: str, dest, follow_symlinks=False):
    is_file(src)
    shutil.copy2(src, dest, follow_symlinks=follow_symlinks)


def copy_match_files(src, dest, regex, follow_src_dir=False, follow_symlinks=False):
    """This api finds files matching the regular expression,
    and copies into the destination directory"""
    mkdir(dest)
    # Find all files with match patern
    files = find_files(regex, src)
    # check for is_file
    for f in files:
        if follow_src_dir:
            relpath = os.path.relpath(f, src)
            dirname = os.path.dirname(relpath)
            dest_dir = os.path.join(dest, dirname)
            mkdir(dest_dir)
            shutil.copy2(f, dest_dir, follow_symlinks=follow_symlinks)
        else:
            shutil.copy2(f, dest, follow_symlinks=follow_symlinks)


def cp_all_files(
    src: str, dest: str, symlinks: bool = False, ignore: bool = None
) -> None:
    if is_dir(src, False) == True:
        for item in os.listdir(src):
            s = os.path.join(src, item)
            d = os.path.join(dest, item)
            if os.path.isfile(s):
                shutil.copy2(s, d)


def copyDirectory(src: str, dst: str, symlinks: bool = False, ignore=None) -> None:
    if not is_dir(dst):
        mkdir(dst)
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if is_dir(s):
            if is_dir(d):
                copy_tree(s, d, symlinks)
            else:
                shutil.copytree(s, d, symlinks, ignore)
        else:
            shutil.copy2(s, d)


def copy_data(
    src: str, dst: str, symlinks: bool = False, ignore=None, silent_discard: bool = True
) -> None:
    """This api finds if the source is a file or a directory, and copies it accordingly
    to the destination directory"""
    if is_file(src):
        copy_file(src, dst, follow_symlinks=symlinks)
    elif is_dir(src):
        copyDirectory(src, dst, symlinks, ignore)
    else:
        err_msg = f"The path: {src} doesn't exist"
        if not silent_discard:
            log.exception(err_msg)
            raise ValueError(err_msg)
        else:
            log.info(err_msg)


# Read file matching regex lines only
def read_file(file_path, regex):
    valid_lines = []
    with open(os.path.expandvars(file_path), "r") as stream:
        line = stream.readline()
        while line:
            line = str(line).strip("\n")
            # FIX # Support
            e = re.search(regex, line)
            if e:
                e = e.string.split(regex)[1].strip(" ")
                e = os.path.expandvars(e)
                valid_lines.append(str(e))
            line = stream.readline()
    return valid_lines


def export_env(file_path: str) -> None:
    for line in read_file(file_path, "export"):
        # FIXME: Check for #/commented one to not to export it.
        (key, _, value) = line.partition("=")
        value = os.path.expandvars(value).strip('"')
        # FIXME: Check for valid substituion
        os.environ[key] = value


def init_conf(file_path="${ROOT}/conf/init.conf", regex="source"):
    data = read_file(file_path, regex)
    for e in data:
        e = os.path.expandvars(e)


class LoadFromFile(argparse.Action):
    """Allows to read from a yaml file and using argparse to set the attribute values.

    Args:
        argparse (:obj:`Namespace`): It reads the user arguments from the file using argparse library.

    Raises:
        CustomError: Raises Custom error message if file reading fails.
    """

    def __call__(self, parser, namespace, values, option_string=None):

        # Read user config file
        try:
            with values as f:
                contents = f.read()
            data = yaml.safe_load(contents)

            # Set the arguments after reading from the file. It will stop reading after one correct read
            for _, v in data.items():
                for nk, nv in v.items():
                    setattr(namespace, nk, nv)
                break
        except Exception:
            detailed_error()


def runcmd(cmd, logfile=None) -> bool:

    ret = True
    if logfile is None:
        try:
            subprocess.check_call(cmd, shell=True)
        except subprocess.CalledProcessError:
            ret = False
    else:
        try:
            subprocess.check_call(cmd, shell=True, stdout=logfile, stderr=logfile)
        except subprocess.CalledProcessError:
            ret = False
    return ret


def check_output(cmd, env=None, silent_discard=False):
    try:
        output = subprocess.check_output(
            cmd, env=env, shell=True, universal_newlines=True, stderr=subprocess.PIPE
        )
        returncode = 0
    except subprocess.CalledProcessError as e:
        err_msg = f'"{cmd}" failed with {e.returncode}'
        if not silent_discard:
            print(err_msg, file=sys.stderr)
            raise Exception(err_msg) from None
        else:
            print(err_msg)
        output = e.output
        returncode = e.returncode
    return output, returncode


def runcmd_p(cmd, log, env=None, cwd="./", expected_code=0, return_code=False):
    """Sends command on to the console and raises assertion error if the
    command is not executed successfully.

    Args:
        cmd (str): Command to be executed.
        log (logger): logger to record debug prints
        cwd (str): PWD path to execute cmd
        expected_code (int): Expected return code value from the function
        return_code (boolean): True = returns cmd exit code, False = Asserts error
    Raises:
        Assertion error only when command execution fails
    """
    log.debug(f"cmd: {cmd}")
    # FIXME: stderror will also get into stdout.
    # Needed to find a way to seperate them.
    with subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
        env=env,
        cwd=cwd,
        encoding="utf-8",
        shell=True,
        bufsize=1,
        executable="/bin/bash",
    ) as p:
        for line in p.stdout:  # b'\n'-separated lines
            log.debug(line)
    retcode = p.wait()
    if retcode != expected_code:
        err_msg = f"{cmd} failed with return code {retcode}, expected {expected_code}"
        log.error(err_msg)
        if return_code:
            return retcode
        else:
            assert False, err_msg
    elif return_code:
        return retcode


# Function to add newline in file.
def add_newline(File, newline):
    with open(File, "a") as fd:
        fd.write(newline + "\n")


# Function to remove line with matched string
def remove_line(File, match_string) -> None:
    with open(File, "r+") as f:
        new_f = f.readlines()
        f.seek(0)
        for line in new_f:
            if match_string not in line:
                f.write(line)
        f.truncate()


# Function to replace string.
def replace_string(File, search_string: str, replace_string: str) -> None:
    if is_file(File) == True:
        with open(File, "r+") as fd:
            filedata = fd.read()
            filedata = re.sub(search_string, replace_string, filedata)
            fd.seek(0)
            fd.write(filedata)
            fd.truncate()
    else:
        err_msg = f"No such file exists: {File}"
        log.error(err_msg)
        raise FileNotFoundError(err_msg)


# Function to replace line if match string found.
def replace_line(File: str, search_string: str, add_line: str) -> None:
    if is_file(File) == True:
        add_line = add_line + "\n"
        with fileinput.input(File, inplace=True) as fd:
            for line in fd:
                if search_string in line:
                    line = add_line
                sys.stdout.write(line)
    else:
        err_msg = f"No such file exists: {File}"
        log.error(err_msg)
        raise FileNotFoundError(err_msg)


def git_clone(
    repo_url, clone_path, repo_branch, rev=None, sparse_checkout_path=[], **kwargs
):
    """Allows to clone git repo with submodules

    Args:
        repo_url: git repo url.
        clone_path: path to clone repo with some name
        repo_branch: git repo branch

    Raises:
        Error: Raises error when failed to clone repo.
    """

    try:
        no_checkout = True if len(sparse_checkout_path) else False
        git.Repo.clone_from(
            repo_url, clone_path, branch=repo_branch, no_checkout=no_checkout, **kwargs
        )
        repo = git.Repo(clone_path)
        if len(sparse_checkout_path):
            with repo.config_writer() as cw:
                cw.set_value("core", "sparseCheckout", "true")
            checkout_path = os.path.join(clone_path, ".git/info/sparse-checkout")
            with open(checkout_path, "w") as fd:
                fd.writelines(sparse_checkout_path)
            repo.git.checkout(repo_branch)

        for submodule in repo.submodules:
            submodule.update(init=True)
            sub_repo = submodule.module()
            sub_repo.git.pull()
        if rev:
            repo.git.checkout(rev)
        print(f"Info: {repo_url} Cloned successfully")
        return True
    except Exception:
        print(f"Error: Failed to clone {repo_url}")
        raise GitError(f"Failed to clone {repo_url}")


def has_key(dict_name, key) -> bool:
    """To check if key available in dictionary
    And checks if key is None

    Args:
       dict_name: Dictionary name
       key: Key name
    """

    warn(
        "has_key is deprecated; use membership test operation `in` instead.",
        DeprecationWarning,
    )

    if isinstance(key, str) and "." in key:
        key = key.split(".")
    else:
        first, rest = key, None

    if isinstance(key, list):
        first, rest = key[0], key[1:]

    if rest:
        return has_key(dict_name[first], rest)
    else:
        if first in dict_name:
            if dict_name[first]:
                return True
            else:
                return False
        else:
            return False


def get_var(config, key: str, raise_except: bool = False) -> Any:
    """This api returns vale of key in config.
    If key is not present, it will return False if raise_except is False.
    else it will raise exception.

    Args:
        config: Configuration from test case
        key: key for which value needed to be returned
        raise_except: if key is not found in config, returning False or Raising
                      exception depends on this flag.

    Returns:
        string: if key is present
    """

    warn(
        "get_var is deprecated; use dictionary .get() method instead.",
        DeprecationWarning,
    )

    if not has_key(config, key):
        if raise_except:
            raise Exception(f"Error: Key not found!")
        else:
            return None
    else:
        if isinstance(key, str) and "." in key:
            key = key.split(".")
        else:
            first, rest = key, None

        if isinstance(key, list):
            first, rest = key[0], key[1:]
        if rest:
            # if rest is not empty, run the function recursively
            return get_var(config[first], rest)
        else:
            return config[first]


def set_var(config, key, value):

    if isinstance(key, str) and "." in key:
        keys = key.split(".")
        tree_dict = value
        for ikey in reversed(keys):
            tree_dict = {ikey: tree_dict}
    else:
        tree_dict = {key: value}

    dict_update(config, tree_dict)


def isvalid_config(config, key, default, valid_list, logger=None):
    """Function that checks if a value for key exist in config and if it doesnt
    exist uses the default value, and see if the value/default exists in the valid_list

    Args:
        config (dict): config object
        key (str): key to used to get value from config
        default (str): default value to be assigned if not value exist for key
        valid_list (list): list of valid choices for key
        logger (logger.getlogger): log the error output. Defaults to None.

    Returns:
        value/False: returns the valid value or False
    """
    val = config.get(key, default)
    if val in valid_list:
        return val
    else:
        if logger is not None:
            logger.error(
                f"User input {key} = {val} is invalid!\n" f"Use values in: {valid_list}"
            )
        return False


def convert_list(*argv):
    """This api takes list/strings and
    returns list
    """
    ret_list = []
    for arg in argv:
        if arg:
            if type(arg) == list:
                ret_list += arg
            else:
                ret_list += [arg]

    return ret_list


def remove_all_files(file_path: str) -> None:
    import glob

    """ This api removes all files under directory
    """
    files = glob.glob(f"{file_path}/*")
    for f in files:
        os.remove(f)


def symlink(src_file: str, dst_file: str) -> None:
    """This api is to symlink file or directory."""
    print(os.getcwd())
    if os.path.islink(src_file):
        os.unlink(src_file)
    else:
        remove(src_file)
    if not is_file(dst_file) or is_dir(dst_file):
        err_msg = f"{dst_file} file or directory not found to symlink"
        log.error(err_msg)
        raise Exception(err_msg)
    os.symlink(dst_file, src_file)


def get_relpath_tcrepo(fpath):
    """This api takes file path and returns
    relative path with reference to tcRepo/ dir
    """
    # File relative path
    frpath = os.path.relpath(fpath, f"{os.environ.get('ROOT')}/tcRepo")
    return "/".join([i for i in frpath.split("/") if i])  # Filter out empty elements


def get_test_path_list(fpath):
    """This api takes full path of test dir and splits into list.
    It will have list after tcRepo onwards.
    """
    fpath = get_relpath_tcrepo(fpath)
    fpath_list = fpath.split("/")
    return [x for x in fpath_list if x]  # Filter out empty values


def get_base_name(fpath):
    """This api takes rel path or full path and returns
    base name"""
    return os.path.basename(fpath)


def get_dir_name(fpath):
    """This api takes file path and returns it's directory
    path"""
    return os.path.dirname(fpath)


def get_dirs(fpath: str, exclude_dirs="") -> List:
    """This api takes path and exclude dirs list,
    returns list of dirs available in provided path
    by default it exclude dirs which startswith ('.', '__')
    """

    dirs = [
        str(dirs).strip("./")
        for dirs in os.listdir(fpath)
        if is_dir(os.path.join(fpath, dirs))
        and dirs not in exclude_dirs
        and not dirs.startswith((".", "__"))
    ]
    return dirs


def get_files(directory, extension=None, filename=None, basename=False, abs_path=False):
    """This api takes directory path and
    returns list of files with the given extension or returns list of files
    with same file name and different extension
    """
    files = []
    for f in os.listdir(directory):
        if extension is None and filename:
            if f.startswith(filename + "."):
                files += [f]
        elif f.endswith("." + extension):
            if basename:
                files += [str(f).replace(f".{extension}", "")]
            else:
                files += [f]

    if abs_path:
        abs_files = []
        for f in files:
            abs_files += [get_abs_path(f"{directory}/{f}")]
        return abs_files

    return files


def get_abs_path(fpath):
    """This api takes file path and returns it's absolute
    path"""
    return os.path.abspath(fpath)


def get_original_path(fpath):
    """This api takes file path and returns it's original
    path. It is equivalent to readlink"""
    return os.path.realpath(fpath)


def overrides(conf, var):
    """This api overrides the dictionary which contains same keys"""
    if isinstance(var, list):
        for item in var:
            if item in conf:
                for key, value in conf[item].items():
                    conf[key] = value
    elif var in conf:
        for key, value in conf[var].items():
            conf[key] = value
    return conf


def find_file(search_file: str, search_path: str):
    """This api find the file in sub-directories and returns
    absolute path of file, if file exists"""
    for File in Path(search_path).glob(f"**/{search_file}"):
        return File


def find_files(search_pattern, search_path):
    """This api find the files matching regex directories and returns
    absolute path of files, if file exists"""

    return glob.glob(f"{search_path}/{search_pattern}")


def get_test_params(mydict, keys=[]):
    """This api converts a dict of lists into combinations
    keys and values. This can be used for dicts with
    any number of levels"""
    lists = []
    if type(mydict) is not dict:
        tmp = []
        if type(mydict) is str:
            tmp += [keys + [mydict]]
        else:
            for item in mydict:
                tmp += [keys + [item]]
        return tmp

    for (key, values) in mydict.items():
        if type(key) is tuple:
            for item in key:
                lists += get_test_params(values, keys + [item])
        else:
            lists += get_test_params(values, keys + [key])

    return lists


def check_if_string_in_file(
    file_name: str, string_to_search: str, search_flag: int = 0
) -> bool:
    """Check if any line in the file contains given string"""
    with open(file_name, "r") as read_obj:
        file_data = read_obj.read()
        if re.search(string_to_search, file_data, flags=search_flag):
            return True
    return False


def random_string(string_length=8):
    letters = string.ascii_lowercase
    return "".join(random.choice(letters) for i in range(string_length))


def get_combinations(args):
    for i, arg in enumerate(args):
        if isinstance(arg, str):
            args[i] = [arg]
        elif arg is None:
            args[i] = [arg]
        elif arg == []:
            args[i] = [None]
    return itertools.product(*args)


def flatten_nested_items(dictionary):
    """Flatten a nested_dict.  iterate through nested dictionary
    (with iterkeys() method) and return with nested keys
    flattened into a tuple"""
    if sys.hexversion < 0x03000000:
        keys = dictionary.iterkeys
        keystr = "iterkeys"
    else:
        keys = dictionary.keys
        keystr = "keys"
    for key in keys():
        value = dictionary[key]
        if hasattr(value, keystr):
            for keykey, value in flatten_nested_items(value):
                yield (key,) + keykey, value
        else:
            yield (key,), value


def get_flattened_dict(dictionary, remove_element=None):
    """This api converts a given nested dictionary into
    a flattened one, by converting all the nested keys into
    combinations of tuple keys.

    For Example:
    mydict = {}
    mydict['a']['b'] = 1
    mydict['a']['b']['c'] = 2

    output = get_flattened_dict(mydict)

    output
    {
    ('a', 'b') = 1,
    ('a', 'b', 'c') = 2
    }
    """
    flattened_dict = flatten_nested_items(dictionary)
    mydict = {}
    for key, val in flattened_dict:
        if remove_element in key and remove_element:
            key = list(key)
            key.remove(remove_element)
            key = tuple(key)
        mydict[key] = val
    return mydict


def dict_update(d, *src):
    for u in src:
        for k, v in u.items():
            if isinstance(v, collections.abc.Mapping):
                d[k] = dict_update(d.get(k, {}), v)
            else:
                d[k] = v
    return d


def filter_dict(mydict, scenario_list):
    """This api filters dictionary based on given scenario list. The scenarios in
    the given list are again a list of keys that has to be checked with the
    flattened tuple keys of a nested dict.
    """
    filtered_set = set()
    flat_dict = get_flattened_dict(mydict)
    for scenario in scenario_list:
        scenario = set(scenario)
        for key, val in flat_dict.items():
            key = set(key)
            if scenario.issubset(key):
                filtered_set.update(set(val))
    return list(filtered_set)


def filter_keys(keys):
    """This api is used to take a string or a list of one string and split it
    appropriately to get the list of keys to look for while filtering a
    dictionary. It is first split on basis of semicolon(;) which acts like a
    separator of scenarios given. Next it's split on comma which is the list
    of keys that the scenario mean.
    """

    if keys is None:
        filtered_keys = [[]]
    else:
        if isinstance(keys, str):
            keys = [keys]
        filtered_list = keys[0].split(";")
        filtered_keys = [val.split(",") for val in filtered_list]

    return filtered_keys


def get_config_data(config, key, temp=None):
    if temp is None:
        temp = config
    if isinstance(key, str) and "." in key:
        key = key.split(".")
    else:
        first, rest = key, None

    if isinstance(key, list):
        first, rest = key[0], key[1:]
    if rest:
        # if rest is not empty, run the function recursively
        return get_config_data(config, rest, temp[first])
    else:
        temp_dict = dict(temp)
        return parse_config(config, temp_dict[first])


def is_filesystem_nfs(dir_path):
    cmd = f"df -P -T {dir_path} | tail -n +2 | awk '{{print $2}}'"
    filesystem, _ = check_output(cmd)
    if filesystem.strip() == "nfs":
        return True
    else:
        return False


def parse_config(config, value):
    """Parse config for string interpolation"""

    config["tmp_value"] = value
    return config["tmp_value"]


def colorstr_to_plainstr(string):
    """Conversion from colour string to plain string if any"""
    ansi_escape = re.compile(r"\x1b(\[.*?[@-~]|\].*?(\x07|\x1b\\))")
    format_string = ansi_escape.sub("", str(string))
    return format_string


def rsync(console, src, dest, exclude_list=[".git*"], timeout=200):
    cmd = f"rsync -aqv {src} {dest}"

    if len(exclude_list) > 1:
        exclude = ",".join(exclude_list)
        cmd += f" --exclude={{{exclude}}}"
    elif len(exclude_list) == 1:
        exclude = "".join(exclude_list)
        cmd += f"  --exclude={exclude}"
    console.runcmd(cmd, timeout=timeout)


def str2bool(value: str) -> bool:
    """Convert string case insensitive True/False to boolean"""
    value = value.lower()
    if value == "true":
        return True
    if value == "false":
        return False


def frange(start, stop=None, step=None):
    """Generate stop inclusive range with decimal step"""
    if stop is None:
        stop = start
        start = 0
    if step is None:
        step = 1
    i = start
    while True:
        if step > 0 and i > stop:
            break
        elif step < 0 and i < stop:
            break
        yield i
        i = round(i + step, 14)  # fix for floating point rounding error


class FileAdapter:
    def __init__(self, logger):
        self.logger = logger
        self._data = ""
        self.ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
        atexit.register(self.exit)

    def write(self, data):
        # NOTE: data can be a partial line, multiple lines
        self._data += data
        lines = re.findall(".*\n", self._data, re.M)  # find lines
        for line in lines:
            self._data = self._data.replace(line, "")
            line = self.ansi_escape.sub("", line)
            self.logger.debug(line.rstrip())

    def flush(self):
        pass  # leave it to logging to flush properly

    def exit(self):
        self.logger.debug(self._data)


def to_nested(dictionary: dict) -> dict:
    """Converts dictionary with dotted keys into nested dictionary

    Args:
        dictionary (dict): Dot keyed dictionary such as Box

    Returns:
        dict: Nested dictionary
    """
    result = {}

    def split_rec(k, v, out):
        # split on "." and call recursively to generate nested dictionaries
        k, *rest = k.split(".", 1)
        if rest:
            split_rec(rest[0], v, out.setdefault(k, {}))
        else:
            out[k] = v

    for k, v in dictionary.items():
        split_rec(k, v, result)
    return result



#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#

"""Provides a shuffled sequence from items in a sequence."""
import collections.abc
from typing import Any, List, Optional, Sequence, Union

from mimesis.providers.base import BaseProvider

__all__ = ["Shuffle"]


class Shuffle(BaseProvider):
    """Class for generating a shuffled sequence from items in a sequence."""

    class Meta:
        """Class for metadata."""

        name = "shuffle"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize attributes.

        :param args: Arguments.
        :param kwargs: Keyword arguments.
        """
        super().__init__(*args, **kwargs)

    def __call__(
        self,
        items: Optional[Sequence[Any]],
        essential: Optional[Sequence[Any]] = None,
        length: int = 0,
    ) -> Sequence[Any]:
        """Generate a shuffled sequence from a sequence.

        Provide a shuffled sequence from the elements in a sequence
        **items**, where selections must include elements in **essential** when
        it is specified. If not specified, a plain shuffle is returned.

        :param items: Non-empty sequence (list, tuple or string) of elements.
        :param essential: Non-empty sequence (list, tuple or string) of elements.
        :param length: Length of sequence (number of elements) to provide.
        :return: Sequence chosen from items.
        :raises TypeError: For non-sequence items.
        :raises ValueError: If negative length.

        >>> from roast.providers.shuffle import Shuffle
        >>> shuffle = Shuffle()
        >>> shuffle(items=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        [8, 4, 2, 10, 7, 6, 9, 3, 1, 5]
        >>> shuffle(items=(1, 2, 3, 4, 5, 6, 7, 8, 9, 10), length=7)
        (1, 5, 10, 8, 2, 4, 6)
        >>> shuffle(items=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10], essential=[3, 6, 9])
        [2, 9, 5, 7, 3, 6, 8]
        >>> shuffle(items=(1, 2, 3, 4, 5, 6, 7, 8, 9, 10), essential=[3, 6, 9], length=5)
        (5, 6, 7, 3, 9)
        """

        if not isinstance(items, collections.abc.Sequence):
            raise TypeError("**items** must be non-empty sequence.")

        if not items:
            raise ValueError("**items** must be a non-empty sequence.")

        data: List = []
        if essential is None:
            if length == 0:
                data = list(items)
                self.random.shuffle(data)
            else:
                data = self.random.sample(items, k=length)
        else:
            if len(set(items)) < length:  # avoid an infinite while loop
                raise ValueError(
                    "There are not enough unique elements in "
                    "**items** to provide the specified **number**."
                )
            for element in essential:
                if element in items:
                    continue
                else:
                    raise ValueError(
                        "All elements in **essential** must exist in **items**."
                    )
            if length == 0:
                length = self.random.randint(len(essential), len(items))
            data = list(essential)
            while len(data) < length:
                item = self.random.choice(items)
                if item not in data:
                    data.append(item)
            self.random.shuffle(data)

        if isinstance(items, list):
            return data
        elif isinstance(items, tuple):
            return tuple(data)
        return "".join(data)



#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#

import os
import re
import sys
import contextlib
import logging
import inspect
from importlib import import_module
from typing import Iterable, List
from config import config, ConfigurationSet, InterpolateEnumType
from roast.utils import *  # pylint: disable=unused-wildcard-import
from roast.utils import overrides as overrides_func

log = logging.getLogger(__name__)


def load_configuration(
    modules: Iterable, interpolate_type=InterpolateEnumType.STANDARD
) -> ConfigurationSet:
    """This function loads configuration files heirarchically. The format types accepted are:
    py, json, ini, yaml, toml

    Args:
        modules (Iterable): configurations

    Returns:
        ConfigurationSet: layered configuration
    """
    cfg = config(
        *modules,
        separator="__",
        ignore_missing_paths=True,
        interpolate=True,
        interpolate_type=interpolate_type,
    )
    return cfg


def get_machine_file(machine: str) -> str:
    machine_file = ""
    if machine:
        module = f"roast.machines.{machine}"
        try:
            m = import_module(module)
            machine_file = inspect.getsourcefile(m)
        except ImportError:
            err_msg = f"{machine} is not a valid machine"
            log.exception(err_msg)
            raise FileNotFoundError(err_msg)
    return machine_file


def generate_conf(
    rootdir: Optional[str] = None,
    test_path: str = "",
    test_name: str = "",
    base_params: Optional[List[str]] = None,
    params: Optional[List[str]] = None,
    overrides: Optional[List[str]] = None,
    machine: str = "",
    interpolate_type=InterpolateEnumType.STANDARD,
) -> ConfigurationSet:
    """Create a :obj:`ConfigurationSet` from a heirarchy of configuration files.

    Args:
        rootdir: The base directory. Defaults to None.
        test_path: Full path of test file. Defaults to empty string.
        test_name: Name for the test. Defaults to empty string.
        base_params: Optional parameters. Defaults to None.
        params: Optional parameters. Defaults to None.
        overrides: Optional key/value pairs to override generated configuration. Defaults to None.
        machine: Optional machine type to override configuration.

    Returns:
        Layered configuration with lowest level taking precedence.
    """

    if rootdir is None:
        rootdir = os.getcwd()
        test_path = inspect.stack()[1][1]
        test_name = inspect.stack()[1][3]
    if base_params is None:
        base_params = []
    if params is None:
        params = []
    if overrides is None:
        overrides = []

    relpath = os.path.relpath(test_path, rootdir)
    relpath_list = relpath.split(os.sep)
    del relpath_list[-1]  # remove file name from list
    if test_name:
        relpath_list.append(test_name)
    relpath_param_list = relpath_list + params

    # Build list of configuration files by iterating each dir level
    supported_extensions = [
        ".py",
        ".ini",
        ".toml",
        ".json",
        ".yaml",
        ".yml",
    ]
    relative_path = ""
    config_list = []

    for relpath_param in [rootdir] + relpath_param_list:  # start with rootdir
        relative_path = os.path.abspath(os.path.join(relative_path, relpath_param))
        search_dir = os.path.join(rootdir, relative_path)
        with contextlib.suppress(FileNotFoundError):
            for filename in os.listdir(search_dir):
                root, ext = os.path.splitext(filename)
                if root == "conf" and ext in supported_extensions:
                    found_conf = os.path.join(search_dir, filename)
                    log.debug(f"Configuration file found at {found_conf}")
                    config_list.append(found_conf)

    file_overrides = []
    var_overrides = []
    subtree_overrides = []
    for override in overrides:
        _, ext = os.path.splitext(override)
        if "=" not in override:
            if ext in supported_extensions:
                file_overrides.append(override)
            else:
                subtree_overrides.append(override)
        elif "=" in override and ext not in supported_extensions:
            var_overrides.append(override)

    # Override files
    for override in file_overrides:
        override_abspath = os.path.abspath(override)
        if is_file(override_abspath):
            log.debug(f"Override configuration file found at {override_abspath}")
            config_list.append(override_abspath)

    # Base configuration parameters
    build_dir = os.path.join(rootdir, "build")
    base_ws_dir = os.path.join(build_dir, *base_params)
    test_base_ws_dir = os.path.join(build_dir, *base_params, *relpath_list)
    ws_dir = os.path.join(build_dir, *base_params, *relpath_list, *params)
    log_dir = os.path.join(ws_dir, "log")
    work_dir = os.path.join(ws_dir, "work")
    images_dir = os.path.join(ws_dir, "images")
    testSuite = (
        relpath_list[0] if test_name else os.path.basename(os.path.normpath(rootdir))
    )

    base_config = {
        "ROOT": rootdir,
        "buildDir": build_dir,
        "testSuite": testSuite,
        "test": test_name,
        "test_path_list": relpath_list,
        "test_param_path_list": relpath_param_list,
        "test_path": os.path.join(rootdir, *relpath_list),
        "base_params": base_params,
        "params": params,
        "wsDir": ws_dir,
        "logDir": log_dir,
        "workDir": work_dir,
        "imagesDir": images_dir,
        "base_ws_dir": base_ws_dir,
        "test_base_ws_dir": test_base_ws_dir,
    }
    config_list.append(base_config)  # type: ignore

    if machine:  # base_params and downstream vars need to be updated
        base_params.insert(0, machine)
        base_ws_dir = os.path.join(build_dir, *base_params)
        test_base_ws_dir = os.path.join(build_dir, *base_params, *relpath_list)
        ws_dir = os.path.join(build_dir, *base_params, *relpath_list, *params)
        log_dir = os.path.join(ws_dir, "log")
        work_dir = os.path.join(ws_dir, "work")
        images_dir = os.path.join(ws_dir, "images")
        subtree_overrides.insert(0, machine)
        machine_config = {
            "machine": machine,
            "base_params": base_params,
            "wsDir": ws_dir,
            "logDir": log_dir,
            "workDir": work_dir,
            "imagesDir": images_dir,
            "base_ws_dir": base_ws_dir,
            "test_base_ws_dir": test_base_ws_dir,
        }
        config_list.append(machine_config)  # type: ignore
        config_list.append(get_machine_file(machine))

    config_list.reverse()  # Lowest level first and takes precedence
    config = load_configuration(config_list, interpolate_type=interpolate_type)

    # append override variable from config to overrides at highest level and takes precedence
    if config.get("overrides"):
        if type(config["overrides"]) is list:
            subtree_overrides = config["overrides"] + subtree_overrides  # type: ignore
        else:
            log.error(
                "Override variable detected but should be of type list not : {}".format(
                    type(config["overrides"])
                )
            )

    # auto override variables in configuration
    if subtree_overrides:
        for override in subtree_overrides:
            overrides_func(config, override)
            log.debug(f"Subtree overrides variable applied: {override}")

    # Override variables
    for override in var_overrides:
        key, value = override.split("=")
        if key in config:
            if type(config.as_dict()[key]) is list:
                value = value.split(",")
        if isinstance(value, str):
            if value.lower() in ["false", "true"]:
                value = str2bool(value)
        log.debug(f"Override variable applied: {key}: {value}")
        config.update({key: value})

    return config



#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#


# Set default logging handler to avoid "No handler found" warnings.
import logging
from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())

# Setup namespace package
from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)  # type: ignore]

from ._version import __version__



#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#

"""Provides a sequence of random choices from items in a sequence."""
import collections.abc
from typing import Any, List, Optional, Sequence, Union

from mimesis.providers.base import BaseProvider

__all__ = ["Choices"]


class Choices(BaseProvider):
    """Class for generating a random sequence from items in a sequence."""

    class Meta:
        """Class for metadata."""

        name = "choices"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize attributes.

        :param args: Arguments.
        :param kwargs: Keyword arguments.
        """
        super().__init__(*args, **kwargs)

    def __call__(
        self,
        items: Optional[Sequence[Any]],
        weights: Optional[Sequence[float]] = None,
        length: int = 1,
        unique: bool = False,
    ) -> Union[Sequence[Any], Any]:
        """Generate a randomly-chosen sequence from a sequence.

        Provide elements randomly chosen from the elements in a sequence
        **items**, where selections are made accoding to the relative weights and
        when **length** is specified the random choices are
        contained in a sequence of the same type of length **length**.
        If **unique** is set to True, constrain a returned sequence to contain
        only unique elements.

        :param items: Non-empty sequence (list, tuple or string) of elements.
        :param weights: Non-empty sequence (list or tuple) of floats.
        :param length: Length of sequence (number of elements) to provide.
        :param unique: If True, ensures provided elements are unique.
        :return: Sequence or uncontained element randomly chosen from items.
        :raises TypeError: For non-sequence items or non-integer length.
        :raises ValueError: If negative length or insufficient unique elements.

        >>> from roast.providers.choices import Choices
        >>> choices = Choices()
        >>> choices(items=['a', 'b', 'c'])
        ['c']
        >>> choices(items=['a', 'b', 'c'], length=1)
        ['a']
        >>> choices(items='abc', length=2)
        'ba'
        >>> choices(items=('a', 'b', 'c'), length=5)
        ('c', 'a', 'a', 'b', 'c')
        >>> choices(items='aabbbccccddddd', length=4, unique=True)
        'cdba'
        >>> choices([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], [8, 3, 9, 2, 10, 1, 8, 3, 9, 2], length=5)
        [2, 8, 5, 7, 5]
        >>> choices([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], [8, 3, 9, 2, 10, 1, 8, 3, 9, 2], length=5, unique=True)
        [5, 9, 7, 3, 1]
        """

        if not isinstance(items, collections.abc.Sequence):
            raise TypeError("**items** must be non-empty sequence.")

        if not items:
            raise ValueError("**items** must be a non-empty sequence.")

        if length < 0:
            raise ValueError("**length** should be a positive integer.")

        data: List[str] = []
        if unique and len(set(items)) < length:  # avoid an infinite while loop
            raise ValueError(
                "There are not enough unique elements in "
                "**items** to provide the specified **number**."
            )
        if unique:
            while len(data) < length:
                (item,) = self.random.choices(items, weights, k=1)
                if unique and item not in data:
                    data.append(item)
        else:
            data = self.random.choices(items, weights, k=length)

        if isinstance(items, list):
            return data
        elif isinstance(items, tuple):
            return tuple(data)
        return "".join(data)



#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#


import os
import sys
import re
from datetime import datetime
import logging
from collections import Counter
from roast.utils import mkdir


def setup_logger(
    log_dirname,
    log_filename=None,
    console_level=logging.INFO,
    file_level=logging.DEBUG,
    console_format="",
    file_format="",
    time_format="",
    report_summary=False,
    search_tokens=None,
    error_tokens=None,
):
    global logger
    logger = Logger(
        log_dirname=log_dirname,
        log_filename=log_filename,
        console_level=console_level,
        file_level=file_level,
        console_format=console_format,
        file_format=file_format,
        time_format=time_format,
        report_summary=report_summary,
        search_tokens=search_tokens,
        error_tokens=error_tokens,
    )
    return logger


def teardown_logger():
    logger.close()


class Logger:
    def __init__(
        self,
        log_dirname,
        log_filename=None,
        console_level=logging.INFO,
        file_level=logging.DEBUG,
        console_format="",
        file_format="",
        time_format="",
        report_summary=False,
        search_tokens=None,
        error_tokens=None,
    ):
        """This defines two logging handlers - file at DEBUG level and console at INFO level.

        Args:
            log_dirname (str): Directory where log files will be written.
            log_filename (str, optional): Name for log file. If not specified,
                name will be generated based on timestamp. Defaults to None.
            console_level (int): Logging level for console. Defaults to logging.INFO or 20.
            file_level (int): Logging level for file. Defaults to logging.DEBUG or 10.
            console_format (str): Format in { style. Defaults to empty string which is {message}.
            file_format (str): Format in { style. Defaults to empty string which is {message}.
            time_format (str): Format in time.strftime(). Defaults to empty string which is %Y-%m-%d %H:%M:%S,uuu.
            report_summary (bool): Write summary to end of log file. Defaults to False,
            search_tokens (list): Tokens summary to add addition to Warning, Error, and Critical tokens.
            error_tokens (list): Error Tokens summary to add additional to the Warning, Error, and Critical tokens
            -- search_tokens and error_token are list of tokens which used for viewing
                the token summary in the respective logs files, it is not used for erroring out.

        """

        mkdir(log_dirname)
        if log_filename is None:
            log_filename = datetime.now().strftime("%Y%m%d-%H%M%S") + ".log"
        self.log_path = os.path.join(log_dirname, log_filename)
        # Root logger
        self.logger = logging.getLogger("roast")
        self.logger.setLevel(logging.DEBUG)
        # File handler
        self.fh = logging.FileHandler(self.log_path)
        self.fh.setLevel(file_level)
        self.time_format = time_format
        self.file_format = file_format
        self.fh.setFormatter(
            logging.Formatter(fmt=file_format, datefmt=time_format, style="{")
        )
        self.logger.addHandler(self.fh)
        # stdout handler
        self.out_h = logging.StreamHandler(sys.stdout)
        self.out_h.setLevel(console_level)
        self.out_h.addFilter(lambda record: record.levelno <= logging.INFO)
        self.out_h.setFormatter(logging.Formatter(console_format, time_format, "{"))
        self.logger.addHandler(self.out_h)
        # stderr handler
        self.err_h = logging.StreamHandler(sys.stderr)
        self.err_h.setLevel(logging.WARNING)
        self.err_h.setFormatter(logging.Formatter(console_format, time_format, "{"))
        self.logger.addHandler(self.err_h)

        self.report_summary = report_summary
        if search_tokens is None:
            search_tokens = []
        self.search_tokens = search_tokens

        if error_tokens is None:
            error_tokens = []
        self.error_tokens = error_tokens

    def write_summary(self):
        """
        Write log summary to the log file
        """

        self.fh.setFormatter(
            logging.Formatter(fmt="{message}", datefmt=self.time_format, style="{")
        )
        results = dict()
        builtin_tokens = ["WARNING", "ERROR", "CRITICAL"]
        tokens = builtin_tokens + self.search_tokens + self.error_tokens
        error_tokens = builtin_tokens + self.error_tokens
        for token in tokens:
            results[token] = []
        with open(self.log_path, "r") as f:
            for line in f:
                for token in tokens:
                    if re.search(rf"{token}", line):
                        results[token].append(line.strip())
        for token, lines in results.items():
            if lines:
                line_data = dict(Counter(lines))

                # Dump logs in logger file  and stdout
                self.logger.info(f"{token} Summary")
                for line, count in line_data.items():
                    self.logger.info(f"{count: >4} {line}")
                    print(f"{count: >4} {line}\n", file=sys.stdout)

                # Dump logs in error file
                if token in error_tokens:
                    print(f"{token} Summary", file=sys.stderr)
                    for line, count in line_data.items():
                        print(f"{count: >4} {line}", file=sys.stderr)

        self.fh.setFormatter(
            logging.Formatter(fmt=self.file_format, datefmt=self.time_format, style="{")
        )

    def close(self):
        if self.report_summary:
            self.write_summary()
        self.logger.removeHandler(self.fh)
        self.logger.removeHandler(self.out_h)
        self.logger.removeHandler(self.err_h)
        self.fh.close()
        self.out_h.close()
        self.err_h.close()



#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#

import logging
import json
import os
import time
import shutil
from roast.utils import is_dir, remove, copyDirectory

log = logging.getLogger(__name__)


def read_json(file_path, decode_numbers=False):
    """function that returns the json dictionary object by reading a json file

    Args:
        file_path (str): path to json file
        decode_numbers (bool): Flag to decode strings to numbers

    Returns:
        dict: retrun the JSON data
    """
    try:
        with open(file_path) as db_file_handle:
            try:
                if decode_numbers:
                    return json.load(db_file_handle, object_hook=_decode_numbers)
                else:
                    return json.load(db_file_handle)
            except ValueError:
                log.error(f"Invalid JSON file {file_path}")
                raise ValueError
            except TypeError:  # NoneType
                log.error(f"Reading {file_path} returned TypeErr")
                raise TypeError
    except IOError:
        log.error(f"Unable to open : {file_path}")
        raise IOError


def write_json(file_path, data, sort=True):
    """function that dictionary object to a file

    Args:
        file_path (str): path to json file to be written to
        data (dict): data to write
        sort (bool, optional): Sort based on keys. Defaults to True.
    """
    try:
        with open(file_path, "w") as db_file_handle:
            json.dump(
                data,
                db_file_handle,
                sort_keys=sort,
                separators=(",", ": "),
                indent=4,
            )
    except IOError:
        log.error(f"Unable to write to : {file_path}")
        raise IOError


def _decode_numbers(object):
    if isinstance(object, str):
        prefix = object[0:2]
        try:
            if prefix in ("0x", "0X"):
                return int(object, 16)
            elif prefix in ("0b", "0B"):
                return int(object, 2)
            elif prefix in ("0o", "0O"):
                return int(object, 8)
            else:
                return float(object)
        except ValueError:
            return object
    elif isinstance(object, dict):
        return {k: _decode_numbers(v) for k, v in object.items()}
    elif isinstance(object, list):
        return [_decode_numbers(v) for v in object]
    else:
        return object


def archive_artifacts(
    dest, copy_workdir=False, repo_path=os.getcwd(), src=None, silent_discard=True
):
    """This method copies images, logs directories from build directory to
    the specified destination path
    Parameters:
        src : source directory to be copied, will be copying build directory
              not mentioned
        dest : destination path
    """
    is_dir_check = False
    if src is None:
        src = os.path.join(repo_path, "build")
    if is_dir(dest):
        remove(dest)
        is_dir_check = True
    try:
        if not copy_workdir:
            copyDirectory(src, dest, ignore=shutil.ignore_patterns("work"))
        else:
            copyDirectory(src, dest)
        if is_dir_check:
            os.system(f"chmod -R 775 {dest}/*")
        else:
            os.system(f"chmod -R 775 {dest}")
    except Exception as err:
        if silent_discard:
            log.debug(err)
        else:
            raise Exception(err) from None


def discard_old_data(path, exclude_files=[], exclude_dirs=[], num_of_days=2):
    """This method cleans the directories if older than the mentioned num of
    days, from the given path
    Parameters:
        path : path from which directories are to be cleaned
        num_of_days : directories or files older than the specified num of
                      days will be deleted
    """
    num_of_days = 86400 * num_of_days
    timestamp = time.time()
    for r, dirs, files in os.walk(path):
        final_dirs, final_files = dirs, files
        for e in exclude_dirs:
            for d in dirs:
                if e in os.path.join(r, d):
                    final_dirs.remove(d)
        for e in exclude_files:
            for f in files:
                if e in os.path.join(r, f):
                    final_files.remove(f)
        for dir in final_dirs:
            try:
                mtime = os.lstat(os.path.join(r, dir)).st_mtime
                if timestamp - num_of_days > mtime:
                    remove(os.path.join(r, dir))
            except Exception as err:
                log.error(err)
        for file in final_files:
            try:
                mtime = os.lstat(os.path.join(r, file)).st_mtime
                if timestamp - num_of_days > mtime:
                    remove(os.path.join(r, file))
            except Exception as err:
                log.error(err)



#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#

from abc import ABCMeta, abstractmethod
from stevedore import driver, named


class ComponentFactory(metaclass=ABCMeta):
    """This is an abstract factory that defines how the two types of components are created."""

    def __init__(self, config):
        self.config = config

    @abstractmethod
    def create_component(self, **args):
        """This creates the component based on component type"""


class TestSuiteFactory(ComponentFactory):
    """This is the concrete class to create the TestSuite component"""

    def __init__(self, config):
        super(TestSuiteFactory, self).__init__(config)

    def create_component(self, component_type):
        ts_mgr = driver.DriverManager(
            namespace="roast.component.testsuite",
            name=component_type,
            invoke_on_load=True,
            invoke_args=(self.config,),
        )
        return ts_mgr


class SystemFactory(ComponentFactory):
    """This is the concrete class to create the System component"""

    def __init__(self, config):
        super(SystemFactory, self).__init__(config)

    def create_component(self, component_types):
        system_mgr = named.NamedExtensionManager(
            namespace="roast.component.system",
            names=component_types,
            invoke_on_load=True,
            invoke_args=(self.config,),
            name_order=True,
        )
        return system_mgr


class Scenario:
    def __init__(self, config):
        self.config = config
        self.testsuite = config.get("roast.testsuite")
        self.system = config.get("roast.system")
        self._ts_mgr = None
        self._system_mgr = None

    @property
    def ts(self):
        return self._ts_mgr.driver

    def sys(self, name: str):
        return self._system_mgr._extensions_by_name[name].obj

    def load_component(self):
        if self.system is not None:
            self._system_mgr = SystemFactory(self.config).create_component(self.system)
        if self.testsuite is not None:
            self._ts_mgr = TestSuiteFactory(self.config).create_component(
                self.testsuite
            )

    def configure_component(self):
        if self.system is not None:
            self._system_mgr.map(configure_extensions, self.config)
        if self.testsuite is not None:
            self._ts_mgr.driver.configure()

    def build_component(self):
        build_result = {}
        if self.system is not None:
            build_result.update(self._system_mgr.map(build_extensions, self.config))
        if self.testsuite is not None:
            build_result[self.testsuite] = self._ts_mgr.driver.build()
        return build_result

    def deploy_component(self):
        deploy_result = {}
        deploy_result[self.testsuite] = self._ts_mgr.driver.deploy()
        return deploy_result

    def run_component(self):
        run_result = {}
        run_result[self.testsuite] = self._ts_mgr.driver.run()
        return run_result


def scenario(config):
    scn = Scenario(config)
    scn.load_component()

    return scn


def configure_extensions(ext, config):
    return (ext.name, ext.obj.configure())


def build_extensions(ext, config):
    return (ext.name, ext.obj.build())



#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#

try:
    # For python 3.8 and later
    import importlib.metadata as importlib_metadata
except ImportError:
    # For everyone else
    import importlib_metadata

import logging
from roast.exceptions import PluginError
from stevedore import ExtensionManager

log = logging.getLogger(__name__)


def register_plugin(name, plugin_type, entry_point) -> None:
    """Registers a plugin dynamically without needing to install as a package.

    Args:
        name (str): Name of plugin to be referenced.
        plugin_type (str): Type to determine plugin namespace.
        entry_point (str): Entry point in the form: some.module:some.attr

    Raises:
        PluginError (Exception): Raised when plugin_type is not supported.
    """
    namespace = ""
    if plugin_type == "system":
        namespace = "roast.component.system"
    elif plugin_type == "testsuite":
        namespace = "roast.component.testsuite"
    elif plugin_type == "serial":
        namespace = "roast.serial"
    elif plugin_type == "board":
        namespace = "roast.board"
    elif plugin_type == "relay":
        namespace = "roast.relay"
    else:
        err_msg = f"Plugin type '{plugin_type}' is not supported"
        log.error(err_msg)
        raise ValueError(err_msg)

    try:
        ep = importlib_metadata.EntryPoint(name, entry_point, namespace)
        e = ExtensionManager(namespace)
        if namespace in e.ENTRY_POINT_CACHE:
            entry_points = e.ENTRY_POINT_CACHE.get(namespace)
            if name not in [entry_point.name for entry_point in entry_points]:
                entry_points.append(ep)
                e.ENTRY_POINT_CACHE[namespace] = entry_points
        else:
            e.ENTRY_POINT_CACHE[namespace] = [ep]
        ep.load()
    except:
        raise PluginError(
            f"Unable to load plugin {name}, {plugin_type}, {entry_point}",
            log_stack=True,
        )



#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#

"""
This lib was implemented with api calls to acquire host & target instances, scp file transffer.
"""

import os
import sys
import time
import pexpect
from pexpect import pxssh
import re
import socket
import logging
from typing import Optional
from roast.utils import convert_list, FileAdapter

log = logging.getLogger(__name__)


def scp_file_transfer(
    self,
    images: Optional[str] = None,
    target_path: Optional[str] = None,
    transfer_to_target: bool = True,
    host_path: Optional[str] = None,
    target_ip: str = "",
    proxy_server: Optional[str] = None,
    user: str = "root",
    password: str = "root",
    timeout: int = 3000,
) -> None:
    """Securely transfers file between host and target.

    Args:
        images: Image files to be transfered. Defaults to None.
        target_path: Path on target. Defaults to None.
        transfer_to_target: Transfer to target if True. Transfer from target if False. Defaults to True.
        host_path: Path on host. Defaults to None.
        target_ip: IP address of target. Defaults to "".
        proxy_server: Address of proxy server. Defaults to None.
        user: Username used for login. Defaults to "root".
        password: Password user for login. Defaults to "root".
        timeout: Time for expected output. Defaults to 3000.

    Raises:
        ConnectionError: When files fail to be transferred in within the timeout window.
    """
    if not images:
        images = f"{self.config['images']}/*"

    cmd = "scp -r -q -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null "
    if proxy_server:
        proxy_cmd = f'-o "ProxyCommand ssh {proxy_server} -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null -W %h:%p" '
        cmd += proxy_cmd

    # Send the file
    if transfer_to_target:
        if not target_path:
            target_path = f"{self.config['target_path']}"
        cmd += f"{images} {user}@{target_ip}:{target_path}"
    # Receive the file
    else:
        if not host_path:
            host_path = f"{self.config['host_path']}"
        cmd += f"{user}@{target_ip}:{images} {host_path}"

    err = f"Failed to scp {images} to {target_path}"
    expected_failures = [
        "lost connection",
        "Name or service not known",
        "Name or service not known",
        "Permission denied",
        "No such file or directory",
        "No route to host",
        "Network is unreachable",
        "No space left on device",
    ]
    expected = ["[Pp]assword", "Do you want to continue", self.hostname]

    if self.prompt:
        expected.append(self.prompt)

    self.sync()
    self.sendline(str(cmd))
    for n in range(5):
        index = self.expect(
            expected_failures,
            expected,
            wait_for_prompt=False,
            err_index=len(convert_list(expected_failures)),
            timeout=float(timeout),
        )
        if index == 0:
            self.sendline(password)
        elif index == 1:
            self.sendline("y")
        elif index in (2, 3):
            return_code = self._exit_non_zero_return(cmd, custom_err="Failed to scp")
            if return_code != 0:
                # Max re-tries exceeded
                if n >= 4:
                    raise ConnectionError(err)
            else:
                log.info(f"scp {images} to {target_path} successfully")
                break
        else:
            raise ConnectionError(err)


def pxssh_login(self, userid: str, password: str) -> None:
    for n in range(5):
        try:
            terminal = pxssh.pxssh()
            terminal.login(self.ip, userid, password)
            terminal.logfile = FileAdapter(log)
            log.info(f"Successfully established ssh connection with {self.ip}")
            self.terminal = terminal
            break
        except pxssh.ExceptionPxssh as e:
            log.info("ssh retry %s of %s \n %s" % (str(n), str(5), str(e)))
        time.sleep(1)
    if n >= 4:
        err = f"Failed to establish ssh connection with {self.ip}"
        log.error(err)
        raise ConnectionError(err)


def ssh_login_user(self, userid: str, password: str) -> None:
    sshcmd = f"ssh -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null {userid}@{self.ip} -y"
    terminal = pexpect.spawn(
        sshcmd, echo=self.echo, encoding="utf-8", codec_errors="replace"
    )
    terminal.logfile = FileAdapter(log)
    self.terminal = terminal
    for n in range(4):
        try:
            index = self.terminal.expect(
                [
                    pexpect.EOF,
                    pexpect.TIMEOUT,
                    "password:",
                    "Permission denied",
                    "(%|#|>|\\$|# )",
                ],
                timeout=60,
            )
            if index == 4:
                break
            elif index == 3:
                raise ConnectionError(f"{sshcmd} with incorrect password: {password}")
            elif index == 2:
                self.terminal.sendline(password)
            elif index == 1:
                raise ConnectionError(f"TIMEOUT wait for password while {sshcmd}")
            else:
                raise ConnectionError(
                    f"EOF while spawn {sshcmd} with password: {password}"
                )
        except ConnectionError as e:
            log.error(e)
        time.sleep(1)

    if n >= 3:
        err = f"Failed to establish ssh connection with {self.ip}, username:{userid} password:{password}"
        log.error(err)
        raise ConnectionError(err)
    self.terminal.sendline("/bin/bash --norc")
    self.terminal.expect("bash-", timeout=60)


def ssh_login(self) -> None:

    if self.hostname == socket.gethostname():
        sshcmd = "/bin/bash --norc"
    else:
        sshcmd = f"{self.cmd} {self.ip} '/bin/bash --norc'"

    terminal = pexpect.spawn(
        sshcmd, echo=self.echo, encoding="utf-8", codec_errors="replace"
    )
    terminal.logfile = FileAdapter(log)

    index = terminal.expect(
        [pexpect.EOF, pexpect.TIMEOUT, "bash-", "password:"], timeout=120
    )
    if index == 2:
        self.terminal = terminal
        if self.non_interactive:
            self.terminal.sendline("bash --norc | cat")
            self.terminal.expect("bash-", timeout=20)
        else:
            self.terminal.sendline("bash --norc")
            self.terminal.expect("bash-", timeout=20)

        self.runcmd(r"PS1='\u@\H:\t:\w\$ '", expected=self.hostname)
        log.info("New console on %s!" % self.ip)
    else:
        if index == 3:
            err = "Password less login not Enabled"
            log.error(err)
        raise ConnectionError(f"Failed to establish ssh connection with {self.ip}")



#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#

from setuptools import setup, find_packages, find_namespace_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with codecs.open(os.path.join(here, *parts), "r") as fp:
        return fp.read()


def get_version_and_cmdclass(package_path):
    """Load version.py module without importing the whole package.

    Template code from miniver
    """
    import os
    from importlib.util import module_from_spec, spec_from_file_location

    spec = spec_from_file_location("version", os.path.join(package_path, "_version.py"))
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    return module.__version__, module.cmdclass


version, cmdclass = get_version_and_cmdclass("roast")

long_description = read("README.md")
packages = find_packages(exclude=["contrib", "docs", "tests"])
packages.extend(
    find_namespace_packages(include=["roast.*"], exclude=["contrib", "docs", "tests"])
)

setup(
    name="roast",
    version=version,
    description="Randomized Okaying Across System Topologies (ROAST) Python Framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Xilinx/roast",
    author="Ching-Hwa Yu",
    author_email="chinghwa@xilinx.com",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
    ],
    keywords="roast",
    license="MIT",
    packages=packages,
    package_data={"": ["*.yaml", "*.tcl"]},
    python_requires=">=3.6, <4",
    install_requires=[
        "python-configuration[yaml,toml]==0.8.2",
        "stevedore>=3.0",
        "gitpython",
        "pexpect",
        "pyyaml",
        "mimesis",
        "filelock",
        "python-box",
        "plusminus",
        "numpy",
        "importlib-metadata<5.0.0",
    ],
    extras_require={
        "dev": [
            "pytest",
            "pytest-mock",
            "pytest-black",
            "pytest-freezegun",
            "miniver",
            "coverage",
        ],
    },
    cmdclass=cmdclass,
)



#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#

import os
import sys
import time
import re
import socket
import inspect
import asyncio
import logging
import atexit
import pexpect
from typing import Optional, Union, List
from roast.ssh import ssh_login_user, ssh_login
from roast.utils import convert_list, colorstr_to_plainstr
from roast.exceptions import ExpectError


class Xexpect:
    """This Xexpect class is wrapper on pexpect"""

    sshcmd = "ssh -t -y -o StrictHostKeyChecking=no -X "

    def __init__(
        self,
        log: logging.Logger,
        hostname: str = socket.gethostname(),
        hostip: Optional[str] = None,
        userid: Optional[str] = None,
        password: Optional[str] = None,
        non_interactive: bool = True,
        exit_nzero_ret: bool = False,
        echo: bool = False,
    ):
        self.log = log
        self.hostname = hostname  # TODO Fix same host running
        self.cmd = self.sshcmd
        self.terminal = None
        self.non_interactive = non_interactive
        self.exit_nzero_ret = exit_nzero_ret  # if set, will assert on non zero returns
        self.echo = echo
        self.timeout_multiplier = 1  # To increase default timeout
        atexit.register(self.exit)
        self._setup_ip_prompt(hostip, hostname)
        self._setup_ssh(userid, password)
        self._setup_init()

    def _setup_ip_prompt(self, hostip, hostname):
        if hostip is None:
            self.ip = hostname
            self.prompt = hostname
        else:
            self.ip = hostip
            self.prompt = "(%|#|>|\\$|# )"

    def _setup_ssh(self, userid, password):
        if None not in (userid, password):
            ssh_login_user(self, userid, password)
        else:
            ssh_login(self)

    def _setup_init(self):
        # Disable History
        self.runcmd("set +o history", expected=self.prompt)

        # Source check_nzero_exit function
        # This function returns return code of a command
        cmd = "check_nzero_exit() { ret=$? ; if [ $ret -ne 0 ]; then echo returncode=$ret; fi ;}"
        self.runcmd(cmd, expected=self.prompt)

    def _exit_non_zero_return(self, cmd, custom_err=None):

        # Send exit status command.
        self.sendline("check_nzero_exit")
        # Expect for prompt
        self.terminal.expect(self.prompt)
        # Search fo return code
        matchObj = re.search(r"returncode=([\d]+)", self.terminal.before)
        if matchObj is not None:
            returncode = matchObj.group(1)
            err_msg = (
                custom_err
                if custom_err
                else f"{cmd} exited with returncode {returncode}"
            )
            self.log.error(err_msg)
            assert False, err_msg
        else:
            index = 0

        return index

    def runcmd(
        self,
        cmd: str,
        expected_failures: Union[None, List[str], str] = None,
        expected: Optional[List[str]] = None,
        wait_for_prompt: bool = True,
        timeout: int = 200,
        err_msg: Optional[str] = None,
        retries: int = 1,
    ) -> int:
        """Sends command on to the console with expected failures and expected output and returns the index of expected string.

        Args:
            cmd: Command to be executed on the console
            expected_failures: List of failure patterns or a single fail pattern string to be expected. Defaults to None.
            expected: List of strings or string. Defaults to None.
            wait_for_prompt: To specify if prompt has to be expected after expected pattern. Defaults to True.
            timeout: Waits for mentioned timeout for the expected string. Defaults to 200.
            err_msg: Error message. Defaults to None.
            retries: Number of retries. Defaults to 1.

        Returns:
            Index of the expected string.
        """

        def _runcmd():
            self.sendline(cmd)
            index = self.expect(
                expected_failures,
                expected,
                wait_for_prompt=wait_for_prompt,
                timeout=timeout,
                err_index=len(convert_list(expected_failures)),
                err_msg=err_msg,
            )

            if self.exit_nzero_ret and not expected:
                index = self._exit_non_zero_return(cmd, custom_err=err_msg)
            return index

        for _ in range(retries - 1):
            try:
                ret = _runcmd()
                break  # break out of loop if success
            except Exception:
                self.log.info("Retrying...")
                time.sleep(10)
        else:
            ret = _runcmd()
        return ret

    def runcmd_list(
        self,
        cmd_list: List[str],
        expected: Optional[List[str]] = None,
        timeout: int = 200,
        err_msg: Optional[str] = None,
        expected_failures: Union[None, List[str], str] = None,
    ) -> None:
        """Sends list of commands to the console and expects the specified string.

        Args:
            cmd_list: List of commands to be run on console.
            expected: Expects for the same string for all the commands. If nothing is specified it expects for prompt. Defaults to None.
            timeout: Waits for mentioned timeout for the expected string. Defaults to 200.
            err_msg: Optional custom error message. Defaults to None.
            expected_failures: List of failure patterns or a single fail pattern string to be expected. Defaults to None.
        """

        for cmd in cmd_list:
            self.runcmd(
                cmd,
                expected_failures=expected_failures,
                expected=expected,
                timeout=timeout,
                err_msg=err_msg,
            )

    def runcmd_async(self, cmd, expected=None, timeout=200):
        self.sendline(cmd)
        return self.expect_async(expected, timeout=timeout)

    def output(self):
        """Returns the output of the previous command executed"""
        return colorstr_to_plainstr(self.terminal.before.rstrip())

    def search(self, srch_str):
        """Takes a regular expression as input and outputs the string that matches the regular expression.

        Args:
            srch_str (str): Regular expression in braces

        Returns:
            str: Matched string.

        Example:
            search("([a-z][A-Z]+)")
        """
        match_obj = re.search(srch_str, self.output())

        if match_obj:
            return match_obj.group(1)
        else:
            return ""

    def send(self, cmd):
        """Sends the command on to the console

        Args:
            cmd (str): cmd to be sent on to the console.
        """
        self.terminal.sendline(cmd)

    def sendline(self, cmd):
        """Sends the command on to the console with linefeed.

        Args:
            cmd (str): cmd to be sent on to the console.
        """
        self.terminal.sendline(cmd)

    def sendcontrol(self, cmd):
        """Sends control characters on to the console.

        Args:
            cmd (str): Control characters to be sent on console.
        """
        self.terminal.sendcontrol(cmd)

    def interact(self):
        # Get Object instance
        """Gives control of the child process to the interactive user."""
        print("Interactive Console:", end="")
        self.terminal.interact()

    def expect(
        self,
        expected_failures: Union[None, List[str], str] = None,
        expected: Optional[List[str]] = None,
        wait_for_prompt: bool = True,
        err_index: int = 0,
        timeout: int = 200,
        err_msg: Optional[str] = None,
    ) -> int:
        """Seeks through the stream until a pattern is matched.

        Args:
            expected_failures: list of failure patterns or a single fail pattern string to be expected. Defaults to None.
            expected: List of patterns or a single pattern string to be expexted. Defaults to None.
            wait_for_prompt: Specify if prompt has to be expected. Defaults to True.
            err_index: Length of expected_failures list. Defaults to 0.
            timeout: Waits for mentioned timeout for the expected string. Defaults to 200.
            err_msg: Custom error message. Defaults to None.

        """

        timeout *= self.timeout_multiplier
        # Default timeout will increase by multiple times, value assigned to timeout_multiplication.

        def _expect(cons, expected, err_index, timeout):
            expected_list = convert_list(pexpect.EOF, pexpect.TIMEOUT, expected)

            err_index += 2
            index = self.terminal.expect(expected_list, timeout=timeout)
            msgs = convert_list(
                "ERROR: Expect returned EOF", "ERROR: Expect returned TIMEOUT", expected
            )
            if index < err_index:
                self.log.error(msgs[index])
                if index == 1:
                    self.log.error(
                        f"Timed out at {timeout}s, while  expecting: {expected}"
                    )
                if err_msg:
                    raise ExpectError(err_msg)
                else:
                    raise ExpectError(msgs[index])
            return index - err_index

        # If expected is not the prompt, there is still info in buffer.
        # It is resulting in log interleaving, we have to expect console
        # to sync buffer
        cons = self
        if expected == None and self.prompt != None:
            expected = self.prompt

        expected_list = convert_list(expected_failures, expected)
        # raises exception when nothing is expected
        if len(expected_list) == 0:
            raise ExpectError("Expected list is empty")

        index = _expect(cons, expected_list, err_index, timeout)
        # Expect again if expected strings are not self.prompt
        if (
            cons.terminal.isalive()
            and wait_for_prompt
            and self.prompt != None
            and self.prompt != expected_list[index + err_index]
        ):
            # Making error index to 0, as prompt is expected
            _expect(cons, self.prompt, 0, timeout)
        return index

    def expect_async(self, expected=None, timeout=200):

        if expected is None:
            expected = self.prompt
        # TODO: mutex lock for coro
        self.coro = self.terminal.expect(
            [pexpect.EOF, pexpect.TIMEOUT, expected], async_=True, timeout=timeout
        )
        return

    def wait(self):
        return asyncio.get_event_loop().run_until_complete(self.coro)

    def sync(self):
        self.runcmd("echo 'sync' | tr '[a-z]' '[A-Z]'", expected=["SYNC"])

    def exit(self):
        if self.terminal:
            self.sendline("exit")
        time.sleep(3)



#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#

"""This breaks up utilies into separate modules but at the same time allows developers to:
from roast.utils import <function>
"""

from .roast_utils import *
from .test_utils import *
from .plugin import *
from .logger import *
from .file import *



#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#

import socket
from abc import ABCMeta, abstractmethod
from stevedore import driver


class BoardBase(metaclass=ABCMeta):
    def __init__(self) -> None:
        self.config = {}
        self.isLive = False
        self.reboot = False
        self.serial = None
        self.xsdb = None
        self.invoke_hwserver = True
        self.invoke_xsdb = True
        self.host_console = None
        self.target_console = None
        self.target_ip = None
        self.host = None
        self.first_boot = True
        self.target_user = "root"
        self.target_password = "root"

    @abstractmethod
    def start(self) -> None:
        """Initiate connection."""

    @abstractmethod
    def put(self, src_file: str, dest_path: str) -> None:
        """Transfer file to target.

        Args:
            src_file (str): Path to file.
            dest_path (str): Target file path.
        """

    @abstractmethod
    def get(self, src_file: str, dest_path: str) -> None:
        """Tranfer file from target.

        Args:
            src_file (str): Target file path.
            dest_path (str): Path to file.
        """

    @abstractmethod
    def reset(self) -> None:
        """Reset or reconnect."""

    def _set_host(self) -> None:
        self.host = self.config.get("remote_host", socket.gethostname())

    def _get_target_ip(self) -> None:
        self.serial.sendline("ifconfig eth0")
        self.serial.expect(expected=["# ", r"\$ "], timeout=10, wait_for_prompt=False)
        self.target_ip = self.serial.search("(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})")


class Board:
    def __init__(self, board_type: str) -> None:
        self._board_mgr = driver.DriverManager(
            namespace="roast.board",
            name=board_type,
            invoke_on_load=True,
        )

    @property
    def driver(self):
        return self._board_mgr.driver

    def start(self) -> None:
        self.driver.start()

    def put(self, src_file, dest_path) -> None:
        self.driver.put(src_file, dest_path)

    def get(self, src_file, dest_path) -> None:
        self.driver.get(src_file, dest_path)

    def reset(self) -> None:
        self.driver.reset()



#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#

from abc import ABCMeta, abstractmethod
from stevedore import driver
from roast.utils import register_plugin


class RelayBase(metaclass=ABCMeta):
    @abstractmethod
    def disconnect(self) -> None:
        """Disconnects relay."""

    @abstractmethod
    def connect(self) -> None:
        """Connects relay."""

    @abstractmethod
    def reconnect(self, seconds: int) -> None:
        """Disconnects for specified time and reconnects.

        Args:
            seconds (int): Amount of time to sleep between disconnect and connect.
        """


class Relay:
    def __init__(self, relay_type: str, **kwargs) -> None:
        if relay_type is None:
            relay_type = "dummy_relay"
            kwargs = {}
            register_plugin(
                "dummy_relay",
                "relay",
                "roast.component.relay:DummyRelay",
            )
        self._relay_mgr = driver.DriverManager(
            namespace="roast.relay",
            name=relay_type,
            invoke_on_load=True,
            invoke_kwds=kwargs,
        )

    @property
    def driver(self):
        return self._relay_mgr.driver

    def disconnect(self) -> None:
        self.driver.disconnect()

    def connect(self) -> None:
        self.driver.connect()

    def reconnect(self, seconds: int = 5) -> None:
        self.driver.reconnect(seconds)


class DummyRelay(RelayBase):
    def __init__(self):
        super().__init__()

    def disconnect(self):
        self.disconnected = True

    def connect(self):
        self.connected = True

    def reconnect(self, seconds: int = 5):
        self.reconnected = True
        self.seconds = seconds



#
# Copyright (c) 2020 Xilinx, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#


import logging

log = logging.getLogger(__name__)


class RoastError(Exception):
    """
    The base exception class for all roast exceptions.
    """

    def __init__(self, message: str = None, log_stack: bool = False) -> None:
        self.message = message or getattr(self.__class__, "message", None)
        super().__init__(message)
        if log_stack:
            log.exception(message)
        else:
            log.error(message)

    def __str__(self):
        return self.message


class DirectoryNotFoundError(RoastError):
    """
    Raised when directory is not found in roast utils.
    """


class GitError(RoastError):
    """
    Raised when a Git error occurs in roast utils.
    """


class ExpectError(RoastError):
    """
    Raised when EOF or TIMEOUT occurs in Xexpect.
    """


class PluginError(RoastError):
    """
    Raised when plugin type is not supported in roast.
    """


class RandomizerError(RoastError):
    """
    Raised when Randomizer failed data generation.
    """
